# -*- coding: utf-8 -*-
#
# File: Poi.py
#
# Copyright (c) 2006 by Copyright (c) 2004 Martin Aspeli
# Generator: ArchGenXML Version 1.5.1-svn
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__author__ = """Martin Aspeli <optilude@gmx.net>"""
__docformat__ = 'plaintext'

import logging

from Products.Archetypes import listTypes
from Products.Archetypes.atapi import process_types
from Products.CMFCore import DirectoryView
from Products.CMFCore import utils as cmfutils
from zope.i18nmessageid import MessageFactory

from Products.Poi.config import ADD_CONTENT_PERMISSIONS
from Products.Poi.config import DEFAULT_ADD_CONTENT_PERMISSION
from Products.Poi.config import PROJECTNAME
from Products.Poi.config import product_globals

logger = logging.getLogger("Poi")
logger.debug('Start initialization of product.')
PoiMessageFactory = MessageFactory('Poi')

DirectoryView.registerDirectory('skins', product_globals)
DirectoryView.registerDirectory('skins/Poi', product_globals)


def initialize(context):
    # imports packages and types for registration
    import interfaces
    import psc
    import content

    interfaces, psc, content  # pyflakes

    # Initialize portal content
    all_content_types, all_constructors, all_ftis = process_types(
        listTypes(PROJECTNAME),
        PROJECTNAME)

    cmfutils.ContentInit(
        PROJECTNAME + ' Content',
        content_types=all_content_types,
        permission=DEFAULT_ADD_CONTENT_PERMISSION,
        extra_constructors=all_constructors,
        fti=all_ftis,
    ).initialize(context)

    # Give it some extra permissions to control them on a per class limit
    for i in range(0, len(all_content_types)):
        klassname = all_content_types[i].__name__
        if klassname not in ADD_CONTENT_PERMISSIONS:
            continue

        context.registerClass(meta_type=all_ftis[i]['meta_type'],
                              constructors=(all_constructors[i],),
                              permission=ADD_CONTENT_PERMISSIONS[klassname])



# -*- coding: utf-8 -*-
from Testing import ZopeTestCase

from Products.Poi.config import DESCRIPTION_LENGTH
from Products.Poi.tests import ptc

default_user = ZopeTestCase.user_name


class TestIssue(ptc.PoiTestCase):
    """Test issue functionality"""

    def afterSetUp(self):
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member2', 'Member Two', 'member2@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member3', 'Member Three', 'member3@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(
            self.folder, 'issue-tracker', managers=('member1', 'member2'))
        self.issue = self.createIssue(self.tracker)

    def testEditIssue(self):
        self.issue.setTitle('title')
        self.issue.setRelease('2.0')
        self.issue.setArea('functionality')
        self.issue.setIssueType('feature')
        self.issue.setSeverity('Critical')
        self.issue.setTargetRelease('2.0')
        self.issue.setDetails('details', mimetype='text/x-web-intelligent')
        self.issue.setSteps('step1\nstep2', mimetype='text/x-web-intelligent')
        # self.issue.setAttachment(None)
        self.issue.setContactEmail('member1@example.com')
        self.issue.setWatchers(('member1', 'member2'), )
        self.issue.setResponsibleManager('member2')

        self.assertEqual(self.issue.Title(), 'title')
        self.assertEqual(self.issue.getRelease(), '2.0')
        self.assertEqual(self.issue.getArea(), 'functionality')
        self.assertEqual(self.issue.getIssueType(), 'feature')
        self.assertEqual(self.issue.getSeverity(), 'Critical')
        self.assertEqual(self.issue.getTargetRelease(), '2.0')
        self.assertEqual(self.issue.getDetails(), 'details')
        self.assertEqual(self.issue.getSteps(), 'step1<br />step2')
        # self.assertEqual(self.issue.getAttachment(), None)
        self.assertEqual(self.issue.getContactEmail(), 'member1@example.com')
        self.assertEqual(self.issue.getWatchers(), ('member1', 'member2'))
        self.assertEqual(self.issue.getWatchers(), ('member1', 'member2'))
        self.assertEqual(self.issue.getResponsibleManager(), 'member2')

    def testIsValid(self):
        self.failUnless(self.issue.isValid())

    def testRenameAfterCreation(self):
        self.failUnless(self.issue.getId() == '1')
        self.createIssue(self.tracker)
        self.failUnless(len(self.tracker.objectIds()) == 2)
        self.failUnless('1' in self.tracker.objectIds())
        self.failUnless('2' in self.tracker.objectIds())

    def testSearchableText(self):
        self.failIf('Response-text' in self.issue.SearchableText())
        self.createResponse(self.issue, text='Response-text')
        self.failUnless('Response-text' in self.issue.SearchableText())

    def testManagersVocabContainsUnassigned(self):
        vocab = self.issue.getManagersVocab()
        self.failUnless('(UNASSIGNED)' in vocab)

    def testReleasesVocabContainsUnassigned(self):
        vocab = self.issue.getManagersVocab()
        self.failUnless('(UNASSIGNED)' in vocab)

    def testValidateWatchers(self):
        self.failUnless(self.issue.validate_watchers(('member1', )) is None)
        self.failIf(self.issue.validate_watchers(('memberX', )) is None)
        self.failIf(self.issue.validate_watchers(('member1', 'memberX', ))
                    is None)

    def testIsWatching(self):
        self.issue.setWatchers(())
        self.failIf(self.issue.isWatching())
        self.issue.setWatchers((default_user, ))
        self.failUnless(self.issue.isWatching())

    def testToggleWatching(self):
        self.failIf(self.issue.isWatching())
        self.issue.toggleWatching()
        self.failUnless(self.issue.isWatching())

    def testTransformDetails(self):
        self.issue.setDetails('Make this a link http://test.com',
                              mimetype='text/x-web-intelligent')
        self.assertEqual(
            self.issue.getDetails(),
            'Make this a link <a href="http://test.com" rel="nofollow">'
            'http://test.com</a>')

    def testTransformSteps(self):
        self.issue.setSteps('Make this a link http://test.com',
                            mimetype='text/x-web-intelligent')
        self.assertEqual(
            self.issue.getSteps(),
            'Make this a link <a href="http://test.com" rel="nofollow">'
            'http://test.com</a>')

    def testExplicitDescription(self):
        self.issue.setDescription('A description')
        self.assertEqual(self.issue.Description(), 'A description')

    def testImplicitDescription(self):
        self.issue.setDetails(
            'A short details section with a link http://test.com',
            mimetype='text/x-web-intelligent')
        self.assertEqual(
            self.issue.Description(),
            'A short details section with a link http://test.com')

    def testImplicitLongDetails(self):
        text = "The quick brown fox jumped over the lazy dog" * 20
        self.issue.setDetails(text, mimetype='text/x-web-intelligent')
        self.assertEqual(self.issue.Description(),
                         text[:DESCRIPTION_LENGTH] + '...')

    def testReadableDescription(self):
        text = ("When pasting html you can get:\r\n    - ugly line breaks,\r\n"
                "- non-breaking spaces.\r\n") * 20
        self.issue.setDetails(text, mimetype='text/x-web-intelligent')
        self.failUnless(self.issue.Description().startswith(
            "When pasting html you can get:\r\n    - ugly line breaks"))
        self.failIf("&nbsp;" in self.issue.Description())
        self.failIf("<br />" in self.issue.Description())

    def testUnicodeDescription(self):
        text = (u"The Japanese call their country Nippon or in their own "
                u"characters (I think): \xe6\x97\xa5\xe6\x9c\xac")
        self.issue.setDetails(text, mimetype='text/x-web-intelligent')
        self.failUnless(self.issue.Description() == text)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestIssue))
    return suite



from zope.interface import Interface


class ITracker(Interface):
    """Marker interface for Poi issue tracker."""


class IIssue(Interface):
    """Marker interface for Poi issue.
    """
    pass


class IResponse(Interface):
    """Marker interface for Poi response.
    """
    pass



import reStructuredText as rst
import textwrap

from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ZODB.POSException import ConflictError
from collective.watcherlist.browser import BaseMail
from collective.watcherlist.utils import su
from collective.watcherlist.utils import get_charset
from zope.component.hooks import getSite
from zope.component import getMultiAdapter
from zope.i18n import translate

from Products.Poi import PoiMessageFactory as _
from Products.Poi.adapters import IResponseContainer
from Products.Poi.interfaces import IIssue
from Products.Poi.interfaces import ITracker

wrapper = textwrap.TextWrapper(initial_indent='    ', subsequent_indent='    ')


class BasePoiMail(BaseMail):

    index = ViewPageTemplateFile('templates/poi_mail.pt')
    plain_index = ''
    css_file_name = 'poi-email.css'

    def plain2rst(self):
        """Try to interpret the plain text as reStructuredText.
        """
        charset = get_charset()
        rstText = self.plain
        ignored, warnings = rst.render(
            rstText, input_encoding=charset, output_encoding=charset)
        if len(warnings.messages) == 0:
            body = rst.HTML(
                rstText, input_encoding=charset, output_encoding=charset)
        else:
            # There are warnings, so we keep it simple.
            body = '<pre>%s</pre>' % rstText
        return body

    @property
    def css(self):
        """Render the styling for the html e-mail.
        """
        if not self.css_file_name:
            return u''
        pps = getMultiAdapter((self.context, self.request),
                              name="plone_portal_state")
        portal = pps.portal()
        if not portal:
            return u''
        try:
            # Render the css, in a way that seems to work for both
            # dtml and plain css files.
            css = '%s' % portal.restrictedTraverse(self.css_file_name)
        except (ConflictError, KeyboardInterrupt):
            raise
        except:
            return u''
        return css

    def options(self):
        # Options to pass to the templates.  We used to return an empty
        # dictionary by default, but we might as well return something useful
        # for most cases.
        context = aq_inner(self.context)
        portal = getSite()
        fromName = portal.getProperty('email_from_name', '')
        portal_membership = getToolByName(portal, 'portal_membership')
        member = portal_membership.getAuthenticatedMember()
        memberInfo = portal_membership.getMemberInfo(member.getUserName())
        stateChanger = member.getUserName()
        if memberInfo:
            stateChanger = memberInfo['fullname'] or stateChanger
        # We expect to be called with an issue as context, but perhaps there
        # are other use cases, like a tracker or a brain or a response.
        # Get hold of the tracker if we can.
        if IIssue.providedBy(context):
            tracker = context.getTracker()
        elif ITracker.providedBy(context):
            tracker = context
        else:
            # This would be strange.
            tracker = None
        if tracker is None:
            tracker_title = ''
        else:
            tracker_title = tracker.title_or_id()
        mapping = dict(
            issue_title=su(context.title_or_id()),
            tracker_title=su(tracker_title),
            response_author=su(stateChanger),
            issue_url=su(context.absolute_url()),
            from_name=su(fromName))
        return mapping

    @property
    def plain(self):
        if not self.plain_index:
            return u''
        output = self.plain_index(**self.options())
        return output

    @property
    def html(self):
        """Render the html version of the e-mail.
        """
        return self.index(**self.options())


class NewIssueMail(BasePoiMail):

    index = ViewPageTemplateFile('templates/poi_email_new_issue_html.pt')
    plain_index = ViewPageTemplateFile(
        'templates/poi_email_new_issue_plain.pt')

    def options(self):
        mapping = super(NewIssueMail, self).options()
        context = aq_inner(self.context)
        portal = getSite()
        portal_membership = getToolByName(portal, 'portal_membership')
        issueCreator = context.Creator()
        issueCreatorInfo = portal_membership.getMemberInfo(issueCreator)
        issueAuthor = issueCreator
        if issueCreatorInfo:
            issueAuthor = issueCreatorInfo['fullname'] or issueCreator

        issueText = context.getDetails(mimetype="text/x-web-intelligent")
        paras = issueText.splitlines()
        issueDetails = '\n\n'.join([wrapper.fill(p) for p in paras])
        mapping['issue_author'] = su(issueAuthor)
        mapping['issue_details'] = su(issueDetails)
        return mapping

    @property
    def subject(self):
        context = aq_inner(self.context)
        tracker = context.getTracker()
        subject = _(
            'poi_email_new_issue_subject_template',
            u"[${tracker_title}] #${issue_id} - New issue: ${issue_title}",
            mapping=dict(
                tracker_title=su(tracker.getExternalTitle()),
                issue_id=su(context.getId()),
                issue_title=su(context.Title())))
        # Make the subject unicode and translate it too.
        subject = su(subject)
        subject = translate(subject, 'Poi', context=self.request)
        return subject


class NewResponseMail(BasePoiMail):

    index = ViewPageTemplateFile('templates/poi_email_new_response_html.pt')
    plain_index = ViewPageTemplateFile(
        'templates/poi_email_new_response_plain.pt')

    def __init__(self, context, request):
        super(NewResponseMail, self).__init__(context, request)
        # With -1 we take the last response, which is a sensible default.
        self.response_id = int(self.request.get('response_id', -1))

    def options(self):
        mapping = super(NewResponseMail, self).options()
        context = aq_inner(self.context)
        folder = IResponseContainer(context)
        response = folder[self.response_id]
        responseText = su(response.text)
        paras = responseText.splitlines()

        # Indent the response details so they are correctly
        # interpreted as a literal block after the double colon behind
        # the 'Response Details' header.  This only really matters
        # when someone interprets this as reStructuredText though.
        responseDetails = u'\n\n'.join([wrapper.fill(p) for p in paras])

        changes = []
        for change in response.changes:
            before = su(change.get('before'))
            after = su(change.get('after'))
            name = su(change.get('name'))
            # Some changes are workflow changes, which can be translated.
            # Note that workflow changes are in the plone domain.
            before = translate(before, 'plone', context=self.request)
            after = translate(after, 'plone', context=self.request)
            name = translate(name, 'Poi', context=self.request)
            changes.append(dict(name=name, before=before, after=after))
        if response.attachment:
            attachment_id = getattr(response.attachment, 'filename', u'')
        else:
            attachment_id = u''

        mapping['response_details'] = responseDetails
        mapping['changes'] = changes
        mapping['attachment_id'] = attachment_id
        return mapping

    @property
    def subject(self):
        context = aq_inner(self.context)
        tracker = context.getTracker()
        subject = _(
            'poi_email_new_response_subject_template',
            u"[${tracker_title}] #${issue_id} - Re: ${issue_title}",
            mapping=dict(
                tracker_title=su(tracker.getExternalTitle()),
                issue_id=su(context.getId()),
                issue_title=su(context.Title())))
        # Ensure that the subject is unicode and translate it too.
        subject = su(subject)
        subject = translate(subject, 'Poi', context=self.request)
        return subject


class ResolvedIssueMail(BasePoiMail):

    index = ViewPageTemplateFile('templates/poi_email_resolved_issue_html.pt')
    plain_index = ViewPageTemplateFile(
        'templates/poi_email_resolved_issue_plain.pt')

    @property
    def subject(self):
        context = aq_inner(self.context)
        tracker = context.getTracker()
        subject = _(
            'poi_email_issue_resolved_subject_template',
            u"[${tracker_title}] Resolved #${issue_id} - ${issue_title}",
            mapping=dict(
                tracker_title=su(tracker.getExternalTitle()),
                issue_id=su(context.getId()),
                issue_title=su(context.Title())))
        # Make the subject unicode and translate it too.
        subject = su(subject)
        subject = translate(subject, 'Poi', context=self.request)
        return subject



import logging

from Products.CMFCore.utils import getToolByName
from collective.watcherlist.interfaces import IWatcherList
from collective.watcherlist.utils import get_member_email

from Products.Poi.interfaces import IIssue

logger = logging.getLogger('Poi')


def add_contact_to_issue_watchers(object, event=None):
    """Add the contactEmail of the issue to the watchers.

    Try to add the userid instead of the email.

    Called when an issue has been initialized or edited.
    """
    value = object.getContactEmail()
    if not value:
        return
    member_email = get_member_email()
    if member_email == value:
        # We can add the userid instead of the email.
        portal_membership = getToolByName(object, 'portal_membership')
        member = portal_membership.getAuthenticatedMember()
        value = member.getId()
    watchers = list(object.getWatchers())
    if value in watchers:
        return
    logger.info('Adding contact %s to watchers of issue %r.', value, object)
    watchers.append(value)
    object.setWatchers(tuple(watchers))


def add_manager_to_issue_watchers(object, event=None):
    """Add manager to issue watchers.

    Add the responsible manager (can be TrackerManager or Technician)
    of the issue to the watchers.

    It might make sense to only do this when the manager is a
    Technician.  Some thoughts about this:

    - It should not matter, as TrackerManagers currently cannot really
      opt out of receiving issue emails, except temporarily until
      someone edits the tracker (see the update_tracker_watchers
      method).

    - But when a user is a TrackerManager, gets assigned an issue, and
      is then made Technician instead, he should still be a watcher.

    - But in any case, when a response is added to the issue, this
      method will get called, so the currently responsible manager
      will be added as watcher.

    """
    manager = object.getResponsibleManager()
    if not manager or manager == '(UNASSIGNED)':
        return
    watchers = list(object.getWatchers())
    if manager in watchers:
        return
    logger.info('Adding manager %s to watchers of issue %r.', manager, object)
    watchers.append(manager)
    object.setWatchers(tuple(watchers))


def update_tracker_watchers(object, event=None):
    """Update tracker watchers.

    If there is a mailing list, make sure it is in the
    extra_addresses.  Okay, this is actually handled automatically by
    the adapter.

    If there is NO mailing list, make sure all tracker managers are
    watchers.

    Note that this means that tracker managers can only temporarily
    unsubscribe: once someone edits the tracker, all tracker managers
    are added again, unless a mailing list has been set.

    """
    watchers = list(object.getWatchers())
    changed = False
    for manager in object.getManagers():
        if manager not in watchers:
            logger.info('Adding manager %s to watchers of tracker %r.',
                        manager, object)
            watchers.append(manager)
            changed = True
    if changed:
        object.setWatchers(tuple(watchers))


def merge_response_changes_to_issue(issue):
    """Update the issue with possible changes due to responses.

    Responses can influence their issue in several ways:

    - The text of the response should be added to the searchable text
      of the issue.

    - The responsible manager may have changed, so the watchers field
      may need to be updated.
    """
    add_manager_to_issue_watchers(issue, event=None)
    issue.reindexObject(idxs=['SearchableText'])
    issue.notifyModified()


def post_issue(object, event):
    """Finalise posting of an issue.

    If an anonymous user is posting, Creator would normally be set to
    the root zope manager, as this user will become the owner.
    Instead we give a more sensible default.

    And we do the 'post' transition.

    And send the initial email.

    """
    portal_membership = getToolByName(object, 'portal_membership')
    if portal_membership.isAnonymousUser():
        object.setCreators(('(anonymous)',))
    add_contact_to_issue_watchers(object, event)
    add_manager_to_issue_watchers(object, event)
    portal_workflow = getToolByName(object, 'portal_workflow')
    portal_workflow.doActionFor(object, 'post')


def mail_issue_change(object, event):
    """Send an email on some transitions of an issue.

    Specifically: new issue and resolved issue.
    """
    if event.transition and event.transition.id == 'post':
        watchers = IWatcherList(object)
        watchers.send('new-issue-mail')
    elif event.new_state.id == 'resolved':
        watchers = IWatcherList(object)
        # Only mail the original poster, if available.
        address = object.getContactEmail()
        if address:
            watchers.send('resolved-issue-mail',
                          only_these_addresses=[address])


def removedResponse(object, event):
    issue = event.oldParent
    if IIssue.providedBy(issue):
        merge_response_changes_to_issue(issue)


def modifiedNewStyleResponse(object, event):
    """A response is modified or created so update its parent.
    """
    if len(event.descriptions) > 0:
        parent = event.descriptions[0]
        if IIssue.providedBy(parent):
            merge_response_changes_to_issue(parent)


def addedNewStyleResponse(object, event):
    """A response has been added.
    """
    issue = event.newParent
    if IIssue.providedBy(issue):
        merge_response_changes_to_issue(issue)
        sendResponseNotificationMail(issue)


def sendResponseNotificationMail(issue):
    # As we take the last response by default, we can keep this simple.
    watchers = IWatcherList(issue)
    watchers.send('new-response-mail')



# -*- coding: utf-8 -*-
#
# File: PoiPscTracker.py
#
# Copyright (c) 2006 by Copyright (c) 2004 Martin Aspeli
# Generator: ArchGenXML Version 1.5.1-svn
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__author__ = """Martin Aspeli <optilude@gmx.net>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from Products.Archetypes.atapi import DisplayList
from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import StringField
from Products.Archetypes.atapi import StringWidget
from Products.Archetypes.atapi import registerType
from Products.CMFCore.utils import getToolByName
from zope.interface import implements
import transaction

from Products.Poi import permissions
from Products.Poi.config import PROJECTNAME
from Products.Poi.config import PSC_TRACKER_ID
from Products.Poi.content.PoiTracker import PoiTracker
from Products.Poi.content.PoiTracker import PoiTracker_schema
from Products.Poi.interfaces import ITracker

schema = Schema((

    StringField(
        name='id',
        widget=StringWidget(
            label="Short name",
            description=("Short name for the tracker - should be 'issues' "
                         "to comply with the standards."),
            label_msgid="Poi_label_psctracker_title",
            description_msgid="Poi_description_psctracker_title",
            i18n_domain='Poi',
        ),
        required=False,
        mode="r"
    ),

),
)

PoiPscTracker_schema = PoiTracker_schema.copy() + schema.copy()
PoiPscTracker_schema['title'].required = False
PoiPscTracker_schema['title'].widget.visible = {'edit': 'invisible',
                                                'view': 'visible'}
del PoiPscTracker_schema['availableReleases']


class PoiPscTracker(PoiTracker):
    """Version of the PoiTracker which supports the
    PloneSoftwareCenter. Intended to be added inside a PSCProject.
    """
    _at_rename_after_creation = True
    archetype_name = 'Issue Tracker'
    implements(ITracker)
    meta_type = 'PoiPscTracker'
    portal_type = 'PoiPscTracker'
    schema = PoiPscTracker_schema
    security = ClassSecurityInfo()

    security.declareProtected(permissions.View, 'getExternalTitle')

    def getExternalTitle(self):
        """ The external title of a PSC tracker is <Project> Issue Tracker."""
        return self.aq_inner.aq_parent.Title() + " Issue Tracker"

    def _renameAfterCreation(self, check_auto_id=False):
        parent = self.aq_inner.aq_parent
        if PSC_TRACKER_ID not in parent.objectIds():
            # Can't rename without a subtransaction commit when using
            # portal_factory!
            transaction.savepoint(optimistic=True)
            self.setId(PSC_TRACKER_ID)

    security.declareProtected(permissions.View, 'getAvailableReleases')

    def getAvailableReleases(self):
        """Get the UIDs of the releases available to the tracker."""
        catalog = getToolByName(self, 'portal_catalog')
        releases = catalog.searchResults(
            portal_type='PSCRelease',
            path='/'.join(self.getPhysicalPath()[:-1]),
            sort_on='getId',
        )
        return [r.UID for r in releases]

    security.declareProtected(permissions.View, 'getReleasesVocab')

    def getReleasesVocab(self):
        """Get the releases available to the tracker as a DisplayList."""
        catalog = getToolByName(self, 'portal_catalog')
        releases = catalog.searchResults(
            portal_type='PSCRelease',
            path='/'.join(self.getPhysicalPath()[:-1]),
            sort_on='getId',
        )
        return DisplayList([(r.UID, r.getId) for r in releases])

    security.declareProtected(permissions.View, 'Title')

    def Title(self):
        """The title of an issue tracker is always "Issue tracker"."""
        return "Issue tracker"


registerType(PoiPscTracker, PROJECTNAME)
# end of class PoiPscTracker



from Products.CMFCore.utils import getToolByName

PROFILE_ID = 'profile-Products.Poi:uninstall'


def uninstall(portal, reinstall=False):
    """Handle some steps that are not done by the default uninstall code.

    When we are reinstalling, some steps must NOT be done, as
    information may get lost then, for example when removing catalog
    indexes or metadata.
    """
    if reinstall:
        return
    setup_tool = getToolByName(portal, 'portal_setup')
    return setup_tool.runAllImportStepsFromProfile(PROFILE_ID)



from Products.Poi.tests import ptc


class TestInstallation(ptc.PoiTestCase):
    """Ensure product is properly installed"""

    def afterSetUp(self):
        self.skins = self.portal.portal_skins
        self.types = self.portal.portal_types
        self.factory = self.portal.portal_factory
        self.catalog = self.portal.portal_catalog
        self.workflow = self.portal.portal_workflow
        self.properties = self.portal.portal_properties
        self.transforms = self.portal.portal_transforms
        self.form_controller = self.portal.portal_form_controller

        self.poiTypes = {'PoiTracker': 'poi_tracker_workflow',
                         'PoiPscTracker': 'poi_tracker_workflow',
                         'PoiIssue': 'poi_issue_workflow'}

    def testDataGridFieldInstalled(self):
        self.failUnless('DataGridWidget' in self.skins.objectIds())

    def testAddRemoveWidgetInstalled(self):
        self.failUnless('AddRemoveWidget' in self.skins.objectIds())

    def testPoiInstalled(self):
        self.failUnless('Poi' in self.skins.objectIds())
        for t in self.poiTypes.keys():
            self.failUnless(t in self.types.objectIds())

    def testWorkflowsInstalled(self):
        for k, v in self.poiTypes.items():
            self.failUnless(v in self.workflow.objectIds())
            self.failUnless(self.workflow.getChainForPortalType(k) == (v, ))

    def testIntelligentTextInstalled(self):
        self.failUnless('web_intelligent_plain_text_to_html' in
                        self.transforms.objectIds())

    def testPortalFactorySetup(self):
        for t in self.poiTypes.keys():
            self.failUnless(t in self.factory.getFactoryTypes())

    def testCatalogMetadataInstalled(self):
        self.failUnless('UID' in self.catalog.schema())

    def testParentMetaTypesNotToQuery(self):
        p = self.properties.navtree_properties
        for t in ('PoiTracker', 'PoiPscTracker', 'PoiIssue'):
            self.failUnless(t in p.getProperty('parentMetaTypesNotToQuery'))

    def testReinstall(self):
        """Reinstalling should not empty our indexes.
        """
        # First of all, we want a few indexes in the catalog.
        wanted = ("getRelease", "getArea", "getIssueType", "getSeverity",
                  "getTargetRelease", "getResponsibleManager")
        indexes = self.catalog.indexes()
        for idx in wanted:
            self.failUnless(idx in indexes)

        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(self.folder, 'issue-tracker',
                                          managers=('member1', ))

        def results(**kwargs):
            # Small helper function.
            return len(self.catalog.searchResults(**kwargs))

        # self.createIssue fills in some default already.
        # First check that we have no match in the getArea index.
        self.assertEquals(results(getArea='ui'), 0)
        self.issue = self.createIssue(
            self.tracker, 'an-issue', release='0.1',
            targetRelease='1.0', responsibleManager='member1')

        # After adding we should have a match.
        self.assertEquals(results(getArea='ui'), 1)
        # Same for the other indexes.
        self.assertEquals(results(getIssueType='bug'), 1)
        self.assertEquals(results(getSeverity='Medium'), 1)
        self.assertEquals(results(getRelease='0.1'), 1)
        self.assertEquals(results(getTargetRelease='1.0'), 1)
        self.assertEquals(results(getResponsibleManager='member1'), 1)

        # We should also be able to access the metadata of the brain.
        brain = self.catalog.searchResults(getResponsibleManager='member1')[0]
        self.assertEquals(brain.getArea, 'ui')
        self.assertEquals(brain.getIssueType, 'bug')
        self.assertEquals(brain.getSeverity, 'Medium')
        self.assertEquals(brain.getRelease, '0.1')
        self.assertEquals(brain.getTargetRelease, '1.0')
        self.assertEquals(brain.getResponsibleManager, 'member1')

        # Now we reinstall Poi.
        quickinstaller = self.portal.portal_quickinstaller
        quickinstaller.reinstallProducts(['Poi'])

        # Now we should still have a match.
        self.assertEquals(results(getArea='ui'), 1)
        self.assertEquals(results(getIssueType='bug'), 1)
        self.assertEquals(results(getSeverity='Medium'), 1)
        self.assertEquals(results(getRelease='0.1'), 1)
        self.assertEquals(results(getTargetRelease='1.0'), 1)
        self.assertEquals(results(getResponsibleManager='member1'), 1)

        # We should also still be able to access the metadata of the brain.
        brain = self.catalog.searchResults(getResponsibleManager='member1')[0]
        self.assertEquals(brain.getArea, 'ui')
        self.assertEquals(brain.getIssueType, 'bug')
        self.assertEquals(brain.getSeverity, 'Medium')
        self.assertEquals(brain.getRelease, '0.1')
        self.assertEquals(brain.getTargetRelease, '1.0')
        self.assertEquals(brain.getResponsibleManager, 'member1')


class TestContentCreation(ptc.PoiTestCase):
    """Ensure content types can be created"""

    def testCreateTracker(self):
        self.createTracker(self.folder, 'tracker')

    def testCreateIssue(self):
        self.createTracker(self.folder, 'tracker')
        self.createIssue(self.folder.tracker)

    def testCreateResponse(self):
        self.createTracker(self.folder, 'tracker')
        self.createIssue(self.folder.tracker)
        self.createResponse(self.folder.tracker['1'])


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestInstallation))
    suite.addTest(makeSuite(TestContentCreation))
    return suite



# -*- coding: utf-8 -*-
from Products.Poi.tests import ptc


class TestResponse(ptc.PoiTestCase):
    """Test response functionality"""

    def afterSetUp(self):
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(self.folder, 'issue-tracker',
                                          managers=('member1', ))
        self.issue = self.createIssue(self.tracker, 'an-issue')
        self.response = self.createResponse(self.issue, 'a-response')
        self.workflow = self.portal.portal_workflow

    def testAccentedCharacters(self):
        catalog = self.portal.portal_catalog
        issue = self.createIssue(
            self.tracker,
            title="\xc3\xa9t\xc3\xa9 est belle.",
            details="C'est plus belle \xc3\xa0 Caf\xc3\xa9 Ren\xc3\xa9.")
        found = len(catalog.searchResults(
            portal_type='PoiIssue',
            SearchableText="\xc3\xa9t\xc3\xa9")) >= 1
        self.failUnless(found)
        found = len(catalog.searchResults(
            portal_type='PoiIssue',
            SearchableText="Ren\xc3\xa9")) >= 1
        self.failUnless(found)

        self.createResponse(issue, "In Dutch 'seas' is 'zee\xc3\xabn'")
        # That should show up in the issue.
        found = len(catalog.searchResults(
            portal_type='PoiIssue',
            SearchableText="zee\xc3\xabn")) >= 1
        self.failUnless(found)


class TestKnownIssues(ptc.PoiTestCase):
    """Test bugs with responses"""

    def afterSetUp(self):
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(self.folder, 'issue-tracker',
                                          managers=('member1', ))
        self.issue = self.createIssue(self.tracker, 'an-issue')
        self.response = self.createResponse(self.issue, 'a-response')
        self.catalog = self.portal.portal_catalog

    def testDeleteResponseLeavesStaleDescription(self):
        found = len(self.catalog.searchResults(
            portal_type='PoiIssue', SearchableText='a-response')) >= 1
        self.failUnless(found)

        from Products.Poi.adapters import IResponseContainer
        container = IResponseContainer(self.issue)
        container.delete('0')
        self.failIf('a-response' in self.issue.SearchableText())
        found = len(self.catalog.searchResults(
            portal_type='PoiIssue', SearchableText='a-response')) >= 1
        self.failIf(found, ("OLD ISSUE RAISING ITS HEAD AGAIN: Deleted "
                            "response causes stale issue SearchableText"))


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestResponse))
    suite.addTest(makeSuite(TestKnownIssues))
    return suite



from Products.CMFPlone.utils import safe_unicode
from Products.Poi.config import ISSUE_LINK_TEMPLATE

import re


try:
    from plone.i18n.normalizer.interfaces import \
        IUserPreferredFileNameNormalizer
    FILE_NORMALIZER = True
except ImportError:
    FILE_NORMALIZER = False


def getNumberFromString(linktext):
    """
    Extract the number from a string with a number in it.
    From 'foo666bar' we get '666'.
    (From 'foobar' we probably end up with problems.)
    """
    pattern = re.compile('[1-9][0-9]*')
    res = pattern.search(linktext)
    if res is not None:
        return linktext[res.start(): res.end()]


def linkBugs(text, ids, patterns, base_url='..'):
    """
    Replace patterns with links to other issues in the same tracker.
    """
    for raw in patterns:
        pos = 0
        pattern = re.compile(raw)
        while True:
            res = pattern.search(text, pos)
            if res is None:
                break
            pos = res.start()

            linktext = text[res.start(): res.end()]
            bug = getNumberFromString(linktext)

            if bug is not None and bug in ids:
                link = ISSUE_LINK_TEMPLATE % dict(
                    base_url=base_url,
                    bug=bug,
                    linktext=linktext)
                text = text[0:pos] + link + text[res.end():]
                pos += len(link)
            else:
                pos += 1

    return text


def linkSvn(text, svnUrl, patterns):
    """
    Replace patterns with links to changesets in a repository.
    (What says it has to be svn?)
    """

    if len(svnUrl) == 0:
        return text

    for raw in patterns:
        pos = 0
        pattern = re.compile(raw)
        while True:
            res = pattern.search(text, pos)
            if res is None:
                break

            linktext = text[res.start(): res.end()]
            rev = getNumberFromString(linktext)

            pos = res.start() + 1
            link = '<a href="' + svnUrl % {'rev': rev} + '">' + linktext + \
                '</a>'
            text = text[0: pos - 1] + link + text[res.end():]
            pos += len(link)

    return text


def normalize_filename(filename, request):
    if FILE_NORMALIZER:
        # Get something the user can recognise.
        filename = IUserPreferredFileNameNormalizer(request).normalize(
            filename)
    filename = safe_unicode(filename)
    # Also, get something that does not raise a ConstraintNotSatisfiedError
    # when migrating to blob attachments.  Found in live data: a filename that
    # includes a newline...
    filename = filename.replace('\n', ' ').replace('\r', ' ')
    return filename



from Acquisition import aq_inner
from Products.Five.browser import BrowserView
from collective.watcherlist.interfaces import IWatcherList


class WatcherView(BrowserView):

    def __call__(self):
        context = aq_inner(self.context)
        # Old style
        # context.toggleWatching()
        # New style
        watchers = IWatcherList(context)
        watchers.toggle_watching()
        self.request.RESPONSE.redirect(context.absolute_url())

    def is_watching(self):
        context = aq_inner(self.context)
        watchers = IWatcherList(context)
        return watchers.isWatching()



from datetime import datetime

from Acquisition import aq_inner
from DateTime import DateTime
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView

from Products.Poi.adapters import IResponseContainer


def fixDate(date):
    if isinstance(date, basestring):
        # Should not happen, but I have seen it locally, and I am
        # not sure if that was just the result of temporarily
        # wrong code or something more important.
        date = DateTime(date)
    return date


def convertDate(date):
    return datetime(date.year(), date.month(), date.day(), date.hour(),
                    date.minute())


def getEntrySortingKey(entry):
    if not hasattr(entry, 'portal_type'):
        # response
        key = entry.get('date')
    else:
        # issue
        key = entry.created

    if callable(key):
        key = key()

    return key


class LogView(BrowserView):

    @property
    def catalog(self):
        return getToolByName(self.context, 'portal_catalog')

    def getTimeDelta(self, zope_date):
        date = convertDate(zope_date)
        now = datetime.now()

        delta = now - date

        # use largest time unit
        minutes = 0
        hours = 0
        days = delta.days

        if not days:
            hours = delta.seconds / 3600
        if not hours and not days:
            minutes = delta.seconds / 60

        return {'minutes': minutes,
                'hours': hours,
                'days': days}

    def getPrettyName(self, username=None, user=None):
        if not username and not user:
            raise ValueError('Both username and user cannot be empty')

        if username:
            portal_membership = getToolByName(self.context,
                                              'portal_membership')
            user = portal_membership.getMemberById(username)

        if user is None:
            # Anonymous
            return username
        return user.getProperty('fullname') or user.getUserName()

    def getLogEntries(self, count=20):
        context = aq_inner(self.context)
        issuefolder = context.restrictedTraverse('@@issuefolder')
        # First we get the most recently modified issues, which means
        # the most recently added, or the ones with the most recent
        # responses.
        issues = [i.getObject() for i in issuefolder.getFilteredIssues(
            sort_on='modified', sort_limit=count, sort_order='reverse')]

        responses = []
        for issue in issues:
            folder = IResponseContainer(issue)
            for res in list(folder):
                if not res:
                    continue
                item = dict(
                    parent=issue,
                    response=res,
                    date=fixDate(res.date))
                responses.append(item)

        items = responses + issues

        # sort entries
        items.sort(key=getEntrySortingKey, reverse=True)

        results = []
        for item in items[:count]:
            if not hasattr(item, 'portal_type'):
                # Response
                date = item.get('date')
                issue = item.get('parent')
                response = item.get('response')
                data = {'type': 'response',
                        'author': self.getPrettyName(response.creator),
                        'date': date,
                        'timedelta': self.getTimeDelta(date),
                        'changes': response.changes,
                        'issue': issue.title_or_id(),
                        'url': issue.absolute_url(),
                        'text': response.rendered_text}
            else:
                # Issue
                data = {'title': item.title_or_id(),
                        'type': item.portal_type,
                        'author': self.getPrettyName(item.Creator()),
                        'date': item.created(),
                        'url': item.absolute_url(),
                        'timedelta': self.getTimeDelta(item.created()),
                        'text': item.getTaggedDetails()}

            results.append(data)

        return results



from Acquisition import aq_base
from DateTime import DateTime
from Products.Archetypes.event import ObjectInitializedEvent
from Products.MailHost.interfaces import IMailHost
from Testing import ZopeTestCase
from zope.component import getSiteManager
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent

from Products.Poi.adapters import IResponseContainer

# Make the boring stuff load quietly
ZopeTestCase.installProduct('CMFCore', quiet=1)
ZopeTestCase.installProduct('CMFDefault', quiet=1)
ZopeTestCase.installProduct('CMFCalendar', quiet=1)
ZopeTestCase.installProduct('CMFTopic', quiet=1)
ZopeTestCase.installProduct('DCWorkflow', quiet=1)
ZopeTestCase.installProduct('CMFHelpIcons', quiet=1)
ZopeTestCase.installProduct('CMFQuickInstallerTool', quiet=1)
ZopeTestCase.installProduct('CMFFormController', quiet=1)
ZopeTestCase.installProduct('GroupUserFolder', quiet=1)
ZopeTestCase.installProduct('ZCTextIndex', quiet=1)
ZopeTestCase.installProduct('TextIndexNG2', quiet=1)
ZopeTestCase.installProduct('SecureMailHost', quiet=1)
ZopeTestCase.installProduct('CMFPlone')
ZopeTestCase.installProduct('Archetypes')
ZopeTestCase.installProduct('PortalTransforms', quiet=1)
ZopeTestCase.installProduct('MimetypesRegistry', quiet=1)
ZopeTestCase.installProduct('DataGridField')
ZopeTestCase.installProduct('AddRemoveWidget')

ZopeTestCase.installProduct('Poi')

from Products.PloneTestCase import PloneTestCase
from Products.CMFPlone.tests.utils import MockMailHost
PloneTestCase.setupPloneSite(products=['Poi'])


class PoiTestCase(PloneTestCase.PloneTestCase):

    class Session(dict):

        def set(self, key, value):
            self[key] = value

    def _setup(self):
        PloneTestCase.PloneTestCase._setup(self)
        # Replace normal mailhost with mock mailhost
        self.portal._original_MailHost = self.portal.MailHost
        self.portal.MailHost = mailhost = MockMailHost('MailHost')
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(mailhost, provided=IMailHost)
        # Make sure our mock mailhost does not give a mailhost_warning
        # in the overview-controlpanel.
        mailhost.smtp_host = 'mock'
        self.portal.email_from_address = 'admin@example.com'

        # Setup session (not sure why)
        self.app.REQUEST['SESSION'] = self.Session()

    def _clear(self, call_close_hook=0):
        self.portal.MailHost = self.portal._original_MailHost
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(aq_base(self.portal._original_MailHost),
                           provided=IMailHost)
        PloneTestCase.PloneTestCase._clear(self)

    def addMember(self, username, fullname, email, roles, last_login_time):
        # Taken from CMFPlone/tests/testMemberDataTool
        self.portal.portal_membership.addMember(username, 'secret', roles, [])
        member = self.portal.portal_membership.getMemberById(username)
        member.setMemberProperties(
            {'fullname': fullname, 'email': email,
             'last_login_time': DateTime(last_login_time)})

    def createTracker(
            self, folder, id, title='', description='', helpText='',
            availableAreas=(
                {'id': 'ui', 'title': 'User interface',
                 'description': 'User interface issues'},
                {'id': 'functionality', 'title': 'Functionality',
                 'description': 'Issues with the basic functionality'},
                {'id': 'process', 'title': 'Process',
                 'description':
                 'Issues relating to the development process itself'},
            ),
            availableIssueTypes=(
                {'id': 'bug', 'title': 'Bug',
                 'description': 'Functionality bugs in the software'},
                {'id': 'feature', 'title': 'Feature',
                 'description': 'Suggested features'},
                {'id': 'patch', 'title': 'Patch',
                 'description': 'Patches to the software'},
            ),
            availableSeverities=['Critical', 'Important', 'Medium', 'Low'],
            defaultSeverity='Medium',
            availableReleases=['2.0', '1.0'],
            managers=[],
            sendNotificationEmails=False,
            mailingList=''):
        """Create a new tracker in the given folder"""
        self.setRoles(['Manager'])
        folder.invokeFactory('PoiTracker', id)
        tracker = getattr(folder, id)
        tracker.setTitle(title)
        tracker.setDescription(description)
        tracker.setHelpText(helpText)
        tracker.setAvailableAreas(availableAreas)
        tracker.setAvailableIssueTypes(availableIssueTypes)
        tracker.setAvailableSeverities(availableSeverities)
        tracker.setDefaultSeverity(defaultSeverity)
        tracker.setAvailableReleases(availableReleases)
        tracker.setManagers(managers)
        tracker.setSendNotificationEmails(sendNotificationEmails)
        tracker.setMailingList(mailingList)
        tracker.reindexObject()
        notify(ObjectInitializedEvent(tracker))
        self.setRoles(['Member'])

        return tracker

    def createIssue(self, tracker, title='An issue',
                    details='Something is wrong', release='(UNASSIGNED)',
                    area='ui', issueType='bug', severity='Medium',
                    targetRelease='(UNASSIGNED)',
                    steps='', attachment=None,
                    contactEmail='submitter@example.com',
                    watchers=(),
                    tags=(),
                    responsibleManager='(UNASSIGNED)'):
        """Create an issue in the given tracker, and perform workflow and
        rename-after-creation initialisation"""
        newId = tracker.generateUniqueId('PoiIssue')
        tracker.invokeFactory('PoiIssue', newId)
        issue = getattr(tracker, newId)
        issue.setTitle(title)
        issue.setRelease(release)
        issue.setArea(area)
        issue.setIssueType(issueType)
        issue.setSeverity(severity)
        issue.setTargetRelease(targetRelease)
        issue.setDetails(details)
        issue.setSteps(steps, mimetype='text/x-web-intelligent')
        issue.setAttachment(attachment)
        issue.setWatchers(watchers)
        # This also adds to the watchers:
        issue.setContactEmail(contactEmail)
        issue.setSubject(tags)
        issue.setResponsibleManager(responsibleManager)
        issue._renameAfterCreation()
        issue.reindexObject()
        notify(ObjectInitializedEvent(issue))
        return issue

    def createResponse(self, issue, text='Response text', issueTransition='',
                       newSeverity=None, newTargetRelease=None,
                       newResponsibleManager=None, attachment=None):
        """Create a response to the given tracker, and perform workflow and
        rename-after-creation initialisation"""
        from Products.Poi.browser.response import Create
        request = issue.REQUEST
        request.form['response'] = text
        request.form['transition'] = issueTransition
        if newSeverity is not None:
            request.form['severity'] = newSeverity
        if newTargetRelease is not None:
            request.form['targetRelease'] = newTargetRelease
        if newResponsibleManager is not None:
            request.form['responsibleManager'] = newResponsibleManager
        if attachment is not None:
            request.form['attachment'] = attachment
        create_view = Create(issue, request)
        # A response is created by calling this view:
        create_view()

        container = IResponseContainer(issue)
        id = str(len(container) - 1)
        response = container[id]

        # In tests we need to fire this event manually:
        notify(ObjectModifiedEvent(response))
        return response


class PoiFunctionalTestCase(PoiTestCase, PloneTestCase.FunctionalTestCase):
    pass



from zope.contentprovider.interfaces import ITALNamespaceData
from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import directlyProvides
from zope.viewlet.interfaces import IViewletManager


class IIssueFolderView(Interface):
    """Abstract a PoiTracker into a folder for issues.
    """

    def getFilteredIssues(criteria, **kwargs):
        """Get the contained issues in the given criteria.
        """

    def getIssueSearchQueryString(criteria, **kwargs):
        """Return a query string for an issue query.
        """

    def buildIssueSearchQuery(criteria, **kwargs):
        """Build canonical query for issue search.
        """

    def getMyIssues(openStates, memberId, manager):
        """Get a catalog query result set of my issues.

        So: all issues assigned to or submitted by the current user,
        with review state in openStates.

        If manager is True, you can add more states.
        """

    def getOrphanedIssues(openStates, memberId):
        """Get a catalog query result set of orphaned issues.

        Meaning: all open issues not assigned to anyone and not owned
        by the given user.
        """


class IResponseAdder(IViewletManager):

    mimetype = Attribute("Mime type for response.")
    use_wysiwyg = Attribute("Boolean: Use kupu-like editor.")

    def transitions_for_display():
        """Get the available transitions for this issue.
        """

    def severities_for_display():
        """Get the available severities for this issue.
        """

    def releases_for_display():
        """Get the releases from the project.
        """

    def managers_for_display():
        """Get the tracker managers.
        """

directlyProvides(IResponseAdder, ITALNamespaceData)


class ICreateResponse(Interface):
    pass



from AccessControl import getSecurityManager
from collective.watcherlist.watchers import WatcherList
from DateTime import DateTime
from persistent import Persistent
from persistent.list import PersistentList
from plone.namedfile.interfaces import NotStorable
from plone.namedfile.storages import MAXCHUNKSIZE
from Products.Poi.interfaces import IIssue
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.event import notify
from zope.interface import Attribute
from zope.interface import implements
from zope.interface import Interface
from zope.lifecycleevent import ObjectAddedEvent
from zope.lifecycleevent import ObjectRemovedEvent
from ZPublisher.HTTPRequest import FileUpload

import logging


logger = logging.getLogger('Products.Poi.adapters')


class IssueWatcherList(WatcherList):

    def __get_watchers(self):
        return self.context.getWatchers()

    def __set_watchers(self, v):
        self.context.setWatchers(v)

    watchers = property(__get_watchers, __set_watchers)


class TrackerWatcherList(WatcherList):

    def __get_send_emails(self):
        return self.context.getSendNotificationEmails()

    def __set_send_emails(self, v):
        self.context.setSendNotificationEmails(v)

    send_emails = property(__get_send_emails, __set_send_emails)

    def __get_extra_addresses(self):
        # Return a tuple, not a string!
        return (self.context.getMailingList(), )

    def __set_extra_addresses(self, v):
        if not isinstance(v, basestring):
            # turn tuple or list into string
            v = '.'.join(v)
        self.context.setMailingList(v)

    extra_addresses = property(__get_extra_addresses, __set_extra_addresses)

    def __get_watchers(self):
        return self.context.getWatchers()

    def __set_watchers(self, v):
        self.context.setWatchers(v)

    watchers = property(__get_watchers, __set_watchers)


class IResponseContainer(Interface):
    pass


class IResponse(Interface):

    text = Attribute("Text of this response")
    rendered_text = Attribute("Rendered text (html) for caching")
    changes = Attribute("Changes made to the issue in this response.")
    creator = Attribute("Id of user making this change.")
    date = Attribute("Date (plus time) this response was made.")
    type = Attribute("Type of response (additional/clarification/reply).")
    mimetype = Attribute("Mime type of the response.")
    attachment = Attribute("File attachment.")

    def add_change(id, name, before, after):
        """Add change to the list of changes.
        """


class ResponseContainer(Persistent):

    implements(IResponseContainer)
    adapts(IIssue)
    ANNO_KEY = 'poi.responses'

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(self.context)
        self.__mapping = annotations.get(self.ANNO_KEY, None)
        if self.__mapping is None:
            self.__mapping = PersistentList()
            annotations[self.ANNO_KEY] = self.__mapping

    def __contains__(self, key):
        '''See interface IReadContainer

        Taken from zope.app.container.btree.

        Reimplement this method, since has_key() returns the key if available,
        while we expect True or False.

        >>> c = ResponseContainer()
        >>> "a" in c
        False
        >>> c["a"] = 1
        >>> "a" in c
        True
        >>> "A" in c
        False
        '''
        return key in self.__mapping

    has_key = __contains__

    def __getitem__(self, i):
        i = int(i)
        return self.__mapping.__getitem__(i)

    def __delitem__(self, item):
        self.__mapping.__delitem__(item)

    def __len__(self):
        return self.__mapping.__len__()

    def __setitem__(self, i, y):
        self.__mapping.__setitem__(i, y)

    def append(self, item):
        self.__mapping.append(item)

    def remove(self, id):
        """Remove item 'id' from the list.

        We don't actually remove the item, we just set it to None,
        so that when you edit item 3 out of 3 and someone deletes
        item 2 you are not left in the water.

        Note that we used to get passed a complete item, not an id.
        """
        id = int(id)
        self[id] = None

    def add(self, item):
        if not IResponse.providedBy(item):
            raise ValueError("IResponse interface not provided.")
        self.append(item)
        id = str(len(self))
        event = ObjectAddedEvent(item, newParent=self.context, newName=id)
        notify(event)

    def delete(self, id):
        # We need to fire an ObjectRemovedEvent ourselves here because
        # self[id].__parent__ is not exactly the same as self, which
        # in the end means that __delitem__ does not fire an
        # ObjectRemovedEvent for us.
        #
        # Also, now we can say the oldParent is the issue instead of
        # this adapter.
        event = ObjectRemovedEvent(self[id], oldParent=self.context,
                                   oldName=id)
        self.remove(id)
        notify(event)


class Response(Persistent):

    implements(IResponse)

    def __init__(self, text):
        self.__parent__ = self.__name__ = None
        self.text = text
        self.changes = PersistentList()
        sm = getSecurityManager()
        user = sm.getUser()
        self.creator = user.getId() or '(anonymous)'
        self.date = DateTime()
        self.type = 'additional'
        self.mimetype = ''
        self.rendered_text = None
        self.attachment = None

    def add_change(self, id, name, before, after):
        """Add a new issue change.
        """
        delta = dict(
            id=id,
            name=name,
            before=before,
            after=after)
        self.changes.append(delta)


class EmptyExporter(object):

    def __init__(self, context):
        self.context = context

    def export(self, export_context, subdir, root=False):
        return


class FileUploadStorable(object):
    # Adapted from plone.namedfile.storages.FileUploadStorable.  That one only
    # handles data from zope.publisher.browser.FileUpload.  Poi uses a hand
    # crafted, non-z3c form, so we just get an old-style FileUpload, which
    # means we need our own storage adapter.

    def store(self, data, blob):
        if not isinstance(data, FileUpload):
            raise NotStorable('Could not store data (not of "FileUpload").')

        data.seek(0)

        fp = blob.open('w')
        block = data.read(MAXCHUNKSIZE)
        while block:
            fp.write(block)
            block = data.read(MAXCHUNKSIZE)
        fp.close()



# -*- coding: utf-8 -*-
#
# File: PoiIssue.py
#
# Copyright (c) 2006 by Copyright (c) 2004 Martin Aspeli
# Generator: ArchGenXML Version 1.5.1-svn
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__author__ = """Martin Aspeli <optilude@gmx.net>"""
__docformat__ = 'plaintext'

import logging
import textwrap

from AccessControl import ClassSecurityInfo
from Acquisition import aq_chain
from Products.AddRemoveWidget.AddRemoveWidget import AddRemoveWidget
from Products.Archetypes.atapi import AttributeStorage
from Products.Archetypes.atapi import BaseFolder
from Products.Archetypes.atapi import BaseFolderSchema
from Products.Archetypes.atapi import BaseObject
from Products.Archetypes.atapi import DisplayList
from Products.Archetypes.atapi import FileField
from Products.Archetypes.atapi import FileWidget
from Products.Archetypes.atapi import LinesField
from Products.Archetypes.atapi import LinesWidget
from Products.Archetypes.atapi import RichWidget
from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import SelectionWidget
from Products.Archetypes.atapi import StringField
from Products.Archetypes.atapi import StringWidget
from Products.Archetypes.atapi import TextField
from Products.Archetypes.atapi import registerType
from Products.CMFCore.utils import getToolByName
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.CMFPlone.utils import getSiteEncoding
from Products.CMFPlone.utils import safe_unicode
from collective.watcherlist.utils import get_member_email
from plone.app.blob.field import BlobField
from plone.memoize import instance
from zope.interface import implements
import transaction

from Products.Poi import PoiMessageFactory as _
from Products.Poi import permissions
from Products.Poi.adapters import IResponseContainer
from Products.Poi.config import DEFAULT_ISSUE_MIME_TYPE
from Products.Poi.config import DESCRIPTION_LENGTH
from Products.Poi.config import ISSUE_MIME_TYPES
from Products.Poi.config import PROJECTNAME
from Products.Poi.interfaces import IIssue
from Products.Poi.interfaces import ITracker

wrapper = textwrap.TextWrapper(initial_indent='    ', subsequent_indent='    ')
logger = logging.getLogger('Poi')

PLAIN_MIMETYPES = (
    'text/x-web-intelligent',
    'text/plain',
)

schema = Schema((

    StringField(
        name='id',
        widget=StringWidget(
            visible={'view': 'invisible', 'edit': 'visible'},
            modes=('view', ),
            label=_(u'Poi_label_issue_id',
                    default=u'Id'),
        ),
        mode="r"
    ),

    StringField(
        name='title',
        widget=StringWidget(
            label=_(u"Poi_label_issue_title",
                    default=u"Title"),
            description=_(
                u'Poi_help_issue_title',
                default=(
                    u"Enter a short, descriptive title for the issue. "
                    u"A good title will make it easier for project "
                    u"managers to identify and respond to the issue.")),
        ),
        required=True,
        accessor="Title",
        searchable=True
    ),

    StringField(
        name='release',
        default="(UNASSIGNED)",
        widget=SelectionWidget(
            label=_(u'Poi_label_release',
                    default=u"Version"),
            description=_(u'Poi_help_release',
                          default=(
                              u"Select the version the issue was found in.")),
            condition="object/isUsingReleases",
        ),
        required=True,
        vocabulary='getReleasesVocab'
    ),

    TextField(
        name='details',
        allowable_content_types=ISSUE_MIME_TYPES,
        widget=RichWidget(
            label=_(u'Poi_label_details',
                    default=u"Details"),
            description=_(u'Poi_help_details',
                          default=u"Please provide further details"),
            rows=15,
            allow_file_upload=False,
        ),
        required=True,
        default_content_type=DEFAULT_ISSUE_MIME_TYPE,
        searchable=True,
        default_output_type="text/html"
    ),

    TextField(
        name='steps',
        allowable_content_types=ISSUE_MIME_TYPES,
        widget=RichWidget(
            label=_(u'Poi_label_steps',
                    default=u"Steps to reproduce"),
            description=_(
                u'Poi_help_steps',
                default=(
                    u"If applicable, please provide the steps to "
                    u"reproduce the error or identify the issue, one per "
                    u"line.")),
            rows=6,
            allow_file_upload=False,
        ),
        default_output_type="text/html",
        default_content_type=DEFAULT_ISSUE_MIME_TYPE,
        searchable=True
    ),

    BlobField(
        name='attachment',
        widget=FileWidget(
            label=_(u'Poi_label_attachment',
                    default=u"Attachment"),
            description=_(
                u'Poi_help_attachment',
                default=(
                    u"You may optionally upload a file attachment. Please "
                    u"do not upload unnecessarily large files.")),
        ),
        storage=AttributeStorage(),
        write_permission=permissions.UploadAttachment
    ),

    StringField(
        name='area',
        widget=SelectionWidget(
            label=_(u'Poi_label_area',
                    default=u"Area"),
            description=_(
                u'Poi_help_area',
                default=u"Select the area this issue is relevant to."),
        ),
        enforceVocabulary=True,
        default_method='getDefaultArea',
        vocabulary='getAreasVocab',
        required=True
    ),

    StringField(
        name='issueType',
        widget=SelectionWidget(
            label=_(u'Poi_label_issueType',
                    default=u"Issue type"),
            description=_(
                u'Poi_help_issueType',
                default=u"Select the type of issue."),
        ),
        enforceVocabulary=True,
        default_method='getDefaultIssueType',
        vocabulary='getIssueTypesVocab',
        required=True
    ),

    StringField(
        name='severity',
        widget=SelectionWidget(
            label=_(u'Poi_label_severity',
                    default=u"Severity"),
            description=_(
                u'Poi_help_severity',
                default=u"Select the severity of this issue."),
            format="radio",
        ),
        vocabulary='getAvailableSeverities',
        default_method='getDefaultSeverity',
        required=True,
        write_permission=permissions.ModifyIssueSeverity
    ),

    StringField(
        name='targetRelease',
        widget=SelectionWidget(
            label=_(u'Poi_label_targetRelease',
                    default=u"Target release"),
            description=_(
                u'Poi_help_targetRelease',
                default=u"Release this issue is targetted to be fixed in"),
            condition="object/isUsingReleases",
        ),
        vocabulary='getReleasesVocab',
        default="(UNASSIGNED)",
        required=True,
        write_permission=permissions.ModifyIssueTargetRelease
    ),

    StringField(
        name='responsibleManager',
        widget=SelectionWidget(
            label=_(u'Poi_label_responsibleManager',
                    default=u"Responsible"),
            description=_(
                u'Poi_help_responsibleManager',
                default=(
                    u"Select which manager, if any, is responsible for "
                    u"this issue.")),
        ),
        vocabulary='getManagersVocab',
        default="(UNASSIGNED)",
        required=True,
        write_permission=permissions.ModifyIssueAssignment
    ),

    StringField(
        name='contactEmail',
        validators=('isEmail', ),
        widget=StringWidget(
            label=_(u'Poi_label_contactEmail',
                    default=u"Contact email address"),
            description=_(
                u'Poi_help_contactEmail',
                default=(
                    u"Please provide an email address where you can be "
                    u"contacted for further information or when a "
                    u"resolution is available. Note that your email "
                    u"address will not be displayed to others.")),
        ),
        required=False,
        default_method='getDefaultContactEmail'
    ),

    LinesField(
        name='watchers',
        widget=LinesWidget(
            label=_(u'Poi_label_watchers',
                    default=u"Issue watchers"),
            description=_(
                u'Poi_help_watchers',
                default=(
                    u"Enter the user ids of members who are watching "
                    u"this issue, one per line. E-mail addresses are "
                    u"allowed too. These persons will "
                    u"receive an email when a response is added to the "
                    u"issue. Members can also add themselves as "
                    u"watchers.")),
        ),
        write_permission=permissions.ModifyIssueWatchers
    ),

    LinesField(
        name='subject',
        widget=AddRemoveWidget(
            label=_(u'Poi_label_subject',
                    default=u"Tags"),
            description=_(
                u'Poi_help_subject',
                default=(
                    u"Tags can be used to add arbitrary categorisation to "
                    u"issues. The list below shows existing tags which "
                    u"you can select, or you can add new ones.")),
        ),
        searchable=True,
        vocabulary='getTagsVocab',
        enforceVocabulary=False,
        write_permission=permissions.ModifyIssueTags,
        accessor="Subject"
    ),

))

PoiIssue_schema = BaseFolderSchema.copy() + \
    schema.copy()
PoiIssue_schema.moveField('subject', after='watchers')


class PoiIssue(BaseFolder, BrowserDefaultMixin):
    """The default tracker
    """
    _at_rename_after_creation = True
    archetype_name = 'Issue'
    implements(IIssue)
    meta_type = 'PoiIssue'
    portal_type = 'PoiIssue'
    schema = PoiIssue_schema
    security = ClassSecurityInfo()

    security.declareProtected(permissions.View, 'getAvailableIssueTransitions')

    def getAvailableIssueTransitions(self):
        """
        Get the available transitions for the issue.
        """
        wftool = getToolByName(self, 'portal_workflow')
        transitions = DisplayList()
        transitions.add('', 'No change')
        for tdef in wftool.getTransitionsFor(self):
            transitions.add(tdef['id'], tdef['title_or_id'])
        return transitions

    security.declareProtected(permissions.View, 'toggleWatching')

    def toggleWatching(self):
        """
        Add or remove the current authenticated member from the list of
        watchers.
        """
        portal_membership = getToolByName(self, 'portal_membership')
        if portal_membership.isAnonymousUser():
            return
        member = portal_membership.getAuthenticatedMember()
        memberId = member.getId()
        watchers = list(self.getWatchers())
        if memberId in watchers:
            watchers.remove(memberId)
        else:
            watchers.append(memberId)
        self.setWatchers(tuple(watchers))

    security.declareProtected(permissions.View, 'isWatching')

    def isWatching(self):
        """
        Determine if the current user is watching this issue or not.
        """
        portal_membership = getToolByName(self, 'portal_membership')
        member = portal_membership.getAuthenticatedMember()
        return member.getId() in self.getWatchers()

    security.declareProtected(permissions.View, 'getLastModificationUser')

    def getLastModificationUser(self):
        """
        Get the user id of the user who last modified the issue, either
        by creating, editing or adding a response to it. May return None if
        the user is unknown.
        """
        return getattr(self, '_lastModificationUser', None)

    # Manually created methods

    def getDefaultContactEmail(self):
        """Get the default email address, that of the creating user"""
        return get_member_email()

    def _renameAfterCreation(self, check_auto_id=False):
        parent = self.getTracker()
        maxId = 0
        for id in parent.objectIds():
            try:
                intId = int(id)
                maxId = max(maxId, intId)
            except (TypeError, ValueError):
                pass
        newId = str(maxId + 1)
        # Can't rename without a subtransaction commit when using
        # portal_factory!
        transaction.savepoint(optimistic=True)
        self.setId(newId)

    def Description(self):
        """Return the explicit description or the details.

        If a description is set manually, return that.  Else returns
        the first 200 characters (defined in config.py) of the
        'details' field.

        We must avoid returning a string in one case and a unicode in
        a different case.  So we always return unicode.
        See http://plone.org/products/poi/issues/135
        """
        value = self.getField('description').get(self)
        if not value:
            value = self.getRawDetails()
            if len(value) > DESCRIPTION_LENGTH:
                value = value[:DESCRIPTION_LENGTH] + "..."
        # Always return unicode.
        if not isinstance(value, unicode):
            encoding = getSiteEncoding(self)
            value = safe_unicode(value, encoding)
        return value

    def validate_watchers(self, value):
        """Make sure watchers are actual user ids or email addresses."""
        membership = getToolByName(self, 'portal_membership')
        plone_utils = getToolByName(self, 'plone_utils')
        notFound = []
        for userId in value:
            member = membership.getMemberById(userId)
            if member is None:
                # Maybe an email address
                if not plone_utils.validateSingleEmailAddress(userId):
                    notFound.append(userId)
        if notFound:
            return "The following user ids could not be found: %s" % \
                ','.join(notFound)
        return None

    def getDefaultSeverity(self):
        """Get the default severity for new issues"""
        tracker = self.getTracker()
        return tracker.getDefaultSeverity()

    def getDefaultIssueType(self):
        """Get the default issue type for new issues.

        If there is only one possible issue type, we select this, else
        we select nothing.
        """
        vocab = self.getIssueTypesVocab()
        if len(vocab) == 1:
            return vocab[0]
        return None

    def getDefaultArea(self):
        """Get the default area for new issues.

        If there is only one possible area, we select this, else
        we select nothing.
        """
        vocab = self.getAreasVocab()
        if len(vocab) == 1:
            return vocab[0]
        return None

    security.declarePublic('isValid')

    def isValid(self):
        """Check if the response is valid.

        Meaning: a response has been filled in.
        """
        errors = {}
        self.Schema().validate(self, self.REQUEST, errors, 1, 1)
        if errors:
            return False
        else:
            return True

    security.declareProtected(permissions.View, 'getIssueTypesVocab')

    def getIssueTypesVocab(self):
        """
        Get the issue types available as a DisplayList.
        """
        tracker = self.getTracker()
        field = tracker.getField('availableIssueTypes')
        return field.getAsDisplayList(tracker)

    security.declareProtected(permissions.View, 'getTagsVocab')

    def getTagsVocab(self):
        """
        Get the available areas as a DispayList.
        """
        tracker = self.getTracker()
        tags = tracker.getTagsInUse()
        vocab = DisplayList()
        for t in tags:
            vocab.add(t, t)
        return vocab

    security.declareProtected(permissions.View, 'getReleasesVocab')

    def getReleasesVocab(self):
        """
        Get the vocabulary of available releases, including the item
        (UNASSIGNED) to denote that a release is not yet assigned.
        """
        vocab = DisplayList()
        vocab.add(
            '(UNASSIGNED)', _(
                u"not_assigned", default=u'(Not assigned)'))
        tracker = self.getTracker()
        trackerVocab = tracker.getReleasesVocab()
        for k in trackerVocab.keys():
            vocab.add(k, trackerVocab.getValue(k), trackerVocab.getValue(k))
        return vocab

    def SearchableText(self):
        """Include in the SearchableText the text of all responses"""
        text = BaseObject.SearchableText(self)
        folder = IResponseContainer(self, None)
        if folder is None:
            # Should Not Happen (TM)
            return text
        try:
            text += ' ' + ' '.join([r.text for r in folder if r])
        except UnicodeDecodeError:
            text = text.decode(
                'utf-8') + ' ' + ' '.join([r.text.decode('utf-8') for r in folder if r])
        return text

    def notifyModified(self):
        BaseFolder.notifyModified(self)
        mtool = getToolByName(self, 'portal_membership')
        member = mtool.getAuthenticatedMember()
        self._lastModificationUser = member.getId()

    security.declareProtected(permissions.View, 'getAreasVocab')

    def getAreasVocab(self):
        """
        Get the available areas as a DispayList.
        """
        tracker = self.getTracker()
        field = tracker.getField('availableAreas')
        return field.getAsDisplayList(tracker)

    @instance.clearbefore
    def setDetails(self, *args, **kwargs):
        self.getField('details').set(self, *args, **kwargs)

    @instance.clearbefore
    def setSteps(self, *args, **kwargs):
        self.getField('steps').set(self, *args, **kwargs)

    def _link_detect_text(self, fieldname, **kwargs):
        # perform link detection
        field = self.getField(fieldname)
        text = field.get(self, **kwargs)
        if field.getContentType(self) not in PLAIN_MIMETYPES:
            # Don't touch text/html or similar.
            return text
        tracker = self.getTracker()
        return tracker.linkDetection(text)

    @instance.memoize
    def getTaggedDetails(self, **kwargs):
        return self._link_detect_text('details', **kwargs)

    @instance.memoize
    def getTaggedSteps(self, **kwargs):
        return self._link_detect_text('steps', **kwargs)

    def getTracker(self):
        """Return the tracker.

        This gets around the problem that the aq_parent of an issue
        that is being created is not the tracker, but a temporary
        folder.
        """
        for parent in aq_chain(self):
            if ITracker.providedBy(parent):
                return parent
        raise Exception(
            "Could not find PoiTracker in acquisition chain of %r" %
            self)


registerType(PoiIssue, PROJECTNAME)
# end of class PoiIssue



# -*- coding: utf-8 -*-
#
# File: PoiTracker.py
#
# Copyright (c) 2006 by Copyright (c) 2004 Martin Aspeli
# Generator: ArchGenXML Version 1.5.1-svn
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__author__ = """Martin Aspeli <optilude@gmx.net>"""
__docformat__ = 'plaintext'

from AccessControl import ClassSecurityInfo
from Products.Archetypes.atapi import DisplayList
from Products.CMFCore.utils import getToolByName
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.DataGridField.DataGridField import DataGridField
from Products.DataGridField.DataGridWidget import DataGridWidget
from zope.interface import implements
try:
    from Products.LinguaPlone import public as atapi
    atapi  # pyflakes
except ImportError:
    # No multilingual support
    from Products.Archetypes import atapi

from Products.Poi import PoiMessageFactory as _
from Products.Poi import permissions
from Products.Poi.config import PROJECTNAME
from Products.Poi.config import ISSUE_RECOGNITION_PATTERNS
from Products.Poi.config import REVISION_RECOGNITION_PATTERNS
from Products.Poi.interfaces import ITracker
from Products.Poi.utils import linkBugs
from Products.Poi.utils import linkSvn

schema = atapi.Schema((

    atapi.StringField(
        name='title',
        widget=atapi.StringWidget(
            label=_(u'Poi_label_tracker_title',
                    default=u"Tracker name"),
            description=_(
                u'Poi_help_tracker_title',
                default=u"Enter a descriptive name for this tracker"),
        ),
        required=True,
        accessor="Title",
        searchable=True
    ),

    atapi.TextField(
        name='description',
        widget=atapi.TextAreaWidget(
            label=_(u'Poi_label_tracker_description',
                    default=u"Tracker description"),
            description=_(
                u'Poi_help_tracker_description',
                default=u"Describe the purpose of this tracker"),
        ),
        use_portal_factory="1",
        accessor="Description",
        searchable=True,
        allowable_content_types=('text/plain'),
    ),

    atapi.TextField(
        name='helpText',
        allowable_content_types=('text/plain', 'text/structured', 'text/html',
                                 'application/msword'),
        widget=atapi.RichWidget(
            label=_(u'Poi_label_helpText',
                    default=u"Help text"),
            description=_(
                u'Poi_help_helpText',
                default=(u"Enter any introductory help text you'd like to "
                         u"display on the tracker front page.")),
        ),
        default_output_type='text/html',
        searchable=True
    ),

    DataGridField(
        name='availableAreas',
        default=({'id': 'ui',
                  'title': 'User interface',
                  'description': 'User interface issues'},
                 {'id': 'functionality',
                  'title': 'Functionality',
                  'description': 'Issues with the basic functionality'},
                 {'id': 'process',
                  'title': 'Process',
                  'description':
                  'Issues relating to the development process itself'}),
        widget=DataGridWidget(
            label=_(u'Poi_label_availableAreas',
                    default=u"Areas"),
            description=_(
                u'Poi_help_availableAreas',
                default="Enter the issue topics/areas for this tracker."),
            column_names=('Short name', 'Title', 'Description'),
        ),
        allow_empty_rows=False,
        required=True,
        columns=('id', 'title', 'description',)
    ),

    DataGridField(
        name='availableIssueTypes',
        default=({'id': 'bug',
                  'title': 'Bug',
                  'description': 'Functionality bugs in the software'},
                 {'id': 'feature',
                  'title': 'Feature',
                  'description': 'Suggested features'},
                 {'id': 'patch',
                  'title': 'Patch',
                  'description': 'Patches to the software'}),
        widget=DataGridWidget(
            label=_(u'Poi_label_availableIssueTypes',
                    default=u"Issue types"),
            description=_(u'Poi_help_availableIssueTypes',
                          default=u"Enter the issue types for this tracker."),
            column_names=('Short name', 'Title', 'Description',),
        ),
        allow_empty_rows=False,
        required=True,
        columns=('id', 'title', 'description')
    ),

    atapi.LinesField(
        name='availableSeverities',
        default=['Critical', 'Important', 'Medium', 'Low'],
        widget=atapi.LinesWidget(
            label=_(u'Poi_label_availableSeverities',
                    default=u"Available severities"),
            description=_(
                u'Poi_help_availableSeverities',
                default=(u"Enter the different type of issue severities "
                         u"that should be available, one per line.")),
        ),
        required=True
    ),

    atapi.StringField(
        name='defaultSeverity',
        default='Medium',
        widget=atapi.SelectionWidget(
            label=_(u'Poi_label_defaultSeverity',
                    default=u"Default severity"),
            description=_(
                u'Poi_help_defaultSeverity',
                default=u"Select the default severity for new issues."),
        ),
        enforceVocabulary=True,
        vocabulary='getAvailableSeverities',
        required=True
    ),

    atapi.LinesField(
        name='availableReleases',
        widget=atapi.LinesWidget(
            label=_(u'Poi_label_availableReleases',
                    default=u"Available releases"),
            description=_(
                u'Poi_help_availableReleases',
                default=(
                    u"Enter the releases which issues can be assigned to, "
                    u"one per line. If no releases are entered, issues "
                    u"will not be organised by release.")),
        ),
        required=False
    ),

    atapi.LinesField(
        name='managers',
        widget=atapi.LinesWidget(
            label=_(u'Poi_label_managers',
                    default=u"Tracker managers"),
            description=_(
                u'Poi_help_managers',
                default=(
                    u"Enter the user ids of the users who will be allowed "
                    u"to manage this tracker, one per line.")),
        ),
        default_method="getDefaultManagers"
    ),

    atapi.LinesField(
        name='technicians',
        widget=atapi.LinesWidget(
            label=_(u'Poi_label_technicians',
                    default=u"Technicians"),
            description=_(
                u'Poi_help_technicians',
                default=(
                    u"Enter the user ids of the users who will be "
                    u"responsible for solving the issues, one per line. "
                    u"Note that having only managers and no technicians "
                    u"is fine: managers can solve issues too.")),
        ),
    ),

    atapi.LinesField(
        name='watchers',
        widget=atapi.LinesWidget(
            label=_(u'Poi_label_tracker_watchers',
                    default=u"Tracker watchers"),
            description=_(
                u'Poi_help_tracker_watchers',
                default=(
                    u"Enter the user ids of members who are watching "
                    u"this tracker, one per line. E-mail addresses are "
                    u"allowed too. These persons will receive "
                    u"an email when an issue or response is added to the "
                    u"tracker. Members can also add themselves as "
                    u"watchers.")),
        ),
        write_permission=permissions.ModifyIssueWatchers
    ),

    atapi.BooleanField(
        name='sendNotificationEmails',
        default=True,
        widget=atapi.BooleanWidget(
            label=_(u'Poi_label_sendNotificationEmails',
                    default=u"Send notification emails"),
            description=_(
                u'Poi_help_sendNotificationEmails',
                default=(
                    u"If selected, tracker managers will receive an email "
                    u"each time a new issue or response is posted, and "
                    u"issue submitters will receive an email when there "
                    u"is a new response and when an issue has been "
                    u"resolved, awaiting confirmation. Technicians will "
                    u"get an email when an issue is assigned to them.")),
        ),
    ),

    atapi.StringField(
        name='mailingList',
        widget=atapi.StringWidget(
            label=_(u'Poi_label_mailingList',
                    default=u"Mailing list"),
            description=_(
                u'Poi_help_mailingList',
                default=(
                    u"If given, and if 'Send notification emails' is "
                    u"selected, an email will be sent to this address "
                    u"each time a new issue or response is posted. "
                    u"Managers will receive individual emails as well. "
                    u"If this is not wanted, you may want to make them "
                    u"technician instead.")),
        ),
        required=False,
        validators=('isEmail',)
    ),

    atapi.StringField(
        name='svnUrl',
        widget=atapi.StringWidget(
            label=_(u'Poi_label_svnurl',
                    default=u"URL to SVN"),
            description=_(
                u'Poi_help_svnurl',
                default=(
                    u"Please enter the Url to the related SVN repository, "
                    u"e.g.: "
                    u"http://dev.plone.org/changeset/%(rev)s/collective "
                    u"for products in the Plone collective.")),
            size='90',
        ),
        required=False,
    ),

))

PoiTracker_schema = atapi.BaseBTreeFolderSchema.copy() + \
    schema.copy()


class PoiTracker(atapi.BaseBTreeFolder, BrowserDefaultMixin):
    """The default tracker
    """
    _at_rename_after_creation = True
    archetype_name = 'Issue Tracker'
    implements(ITracker)
    meta_type = 'PoiTracker'
    portal_type = 'PoiTracker'
    schema = PoiTracker_schema
    security = ClassSecurityInfo()

    # Methods

    security.declarePrivate('linkDetection')

    def linkDetection(self, text):
        """
        Detects issues and svn revision tags and creates links.
        """
        # In case we get something not string like, we just return
        # text without change
        if not isinstance(text, basestring):
            return text
        catalog = getToolByName(self, 'portal_catalog')
        issuefolder = self.restrictedTraverse('@@issuefolder')
        issues = catalog.searchResults(issuefolder.buildIssueSearchQuery(None))
        ids = frozenset([issue.id for issue in issues])
        text = linkBugs(text, ids, ISSUE_RECOGNITION_PATTERNS,
                        base_url=self.absolute_url())
        svnUrl = self.getSvnUrl()
        text = linkSvn(text, svnUrl, REVISION_RECOGNITION_PATTERNS)
        return text

    security.declareProtected(permissions.View, 'isUsingReleases')

    def isUsingReleases(self):
        """Return a boolean indicating whether this tracker is using releases.
        """
        return len(self.getAvailableReleases()) > 0

    security.declareProtected(permissions.View, 'getReleasesVocab')

    def getReleasesVocab(self):
        """
        Get the releases available to the tracker as a DisplayList.
        """
        items = self.getAvailableReleases()
        vocab = atapi.DisplayList()
        for item in items:
            vocab.add(item, item)
        return vocab

    security.declareProtected(permissions.View, 'getTagsInUse')

    def getTagsInUse(self):
        """Get a list of the issue tags in use in this tracker."""
        catalog = getToolByName(self, 'portal_catalog')
        issues = catalog.searchResults(portal_type='PoiIssue',
                                       path='/'.join(self.getPhysicalPath()))
        tags = {}
        for i in issues:
            for s in i.Subject:
                tags[s] = 1
        keys = tags.keys()
        keys.sort(lambda x, y: cmp(x.lower(), y.lower()))
        return keys

    security.declareProtected(permissions.View, 'getExternalTitle')

    def getExternalTitle(self):
        """ Get the external title of this tracker.

        This will be the name used in outgoing emails, for example.
        """
        return self.Title()

    # Manually created methods

    def canSelectDefaultPage(self):
        """Explicitly disallow selection of a default-page."""
        return False

    def _updateRolesField(self, field_name, new_values):
        """Update the roles belonging to the field, and set the field value.

        For the 'managers' field: set the list of tracker managers,
        and give them the TrackerManager local role.

        For the 'techniciancs' field: set the list of technicians, and
        give them the Technician local role.

        Also clean up old roles.
        """
        if field_name == 'managers':
            role = 'TrackerManager'
        elif field_name == 'technicians':
            role = 'Technician'
        else:
            raise ValueError("Wrong tracker field name %s" % field_name)

        field = self.getField(field_name)
        current_values = field.get(self)
        field.set(self, new_values)

        to_remove = [m for m in current_values if m not in new_values]
        to_add_or_keep = new_values
        for user_id in to_remove:
            local_roles = list(self.get_local_roles_for_userid(user_id))
            if role in local_roles:
                local_roles.remove(role)
                # manage_setLocalRoles fails when called with zero roles.
                if local_roles:
                    self.manage_setLocalRoles(user_id, local_roles)
                else:
                    self.manage_delLocalRoles(to_remove)
        for user_id in to_add_or_keep:
            local_roles = list(self.get_local_roles_for_userid(user_id))
            if role not in local_roles:
                local_roles.append(role)
                self.manage_setLocalRoles(user_id, local_roles)

    def getManagersVocab(self, strict=False):
        """
        Get the managers available as a DisplayList. The first item is 'None',
        with a key of '(UNASSIGNED)'.

        Note, we now also allow Technicians here, unless we are called
        with 'strict' is True.
        """
        vocab = DisplayList()
        vocab.add(
            '(UNASSIGNED)', _(
                u"not_assigned", default=u'(Not assigned)'))
        mtool = getToolByName(self, 'portal_membership')
        for item in self.getManagers():
            user = mtool.getMemberById(item)
            if user:
                fullname = user.getProperty('fullname', item) or item
            else:
                fullname = item

            vocab.add(item, fullname)
        if not strict:
            for item in self.getTechnicians():
                user = mtool.getMemberById(item)
                if user:
                    fullname = user.getProperty('fullname', item) or item
                else:
                    fullname = item

                vocab.add(item, fullname)
            return vocab

    def getStrictManagersVocab(self):
        """
        Get the managers available as a DisplayList. The first item is 'None',
        with a key of '(UNASSIGNED)'.

        Note, this vocabulary is strictly for TrackerManagers, so not
        for Technicians.  It is not actually used by default, but can
        be handy for third parties.
        """
        return self.getManagersVocab(strict=True)

    security.declareProtected(permissions.ModifyPortalContent, 'setManagers')

    def setManagers(self, managers):
        """Set the list of tracker managers, and give them the
        TrackerManager local role.
        """
        self._updateRolesField('managers', managers)

    security.declareProtected(
        permissions.ModifyPortalContent, 'setTechnicians')

    def setTechnicians(self, technicians):
        """Set the list of technicians, and give them the
        Technician local role.
        """
        self._updateRolesField('technicians', technicians)

    security.declarePublic('getIssueWorkflowStates')

    def getIssueWorkflowStates(self):
        """Get a DisplayList of the workflow states available on issues."""
        portal_workflow = getToolByName(self, 'portal_workflow')
        chain = portal_workflow.getChainForPortalType('PoiIssue')
        workflow = getattr(portal_workflow, chain[0])
        states = getattr(workflow, 'states')
        vocab = atapi.DisplayList()
        for id, state in states.items():
            vocab.add(id, state.title)
        return vocab.sortedByValue()

    def _validate_user_ids(self, user_ids):
        """Make sure the user ids are actual user ids"""
        membership = getToolByName(self, 'portal_membership')
        not_found = []
        for user_id in user_ids:
            member = membership.getMemberById(user_id)
            if member is None:
                not_found.append(user_id)
        if not_found:
            return "The following user ids could not be found: %s" % \
                ','.join(not_found)

    def validate_managers(self, value):
        """Make sure issue tracker managers are actual user ids"""
        return self._validate_user_ids(value)

    def validate_technicians(self, value):
        """Make sure issue technicians are actual user ids"""
        return self._validate_user_ids(value)

    def validate_watchers(self, value):
        """Make sure watchers are actual user ids or email addresses."""
        membership = getToolByName(self, 'portal_membership')
        plone_utils = getToolByName(self, 'plone_utils')
        notFound = []
        for userId in value:
            member = membership.getMemberById(userId)
            if member is None:
                # Maybe an email address
                if not plone_utils.validateSingleEmailAddress(userId):
                    notFound.append(userId)
        if notFound:
            return "The following user ids could not be found: %s" % \
                ','.join(notFound)
        return None

    def getDefaultManagers(self):
        """The default list of managers should include the tracker owner"""
        return (self.Creator(), )


atapi.registerType(PoiTracker, PROJECTNAME)



from Products.Poi.tests import ptc
from Products.Poi.browser.notifications import BasePoiMail


class DummyPoiMail(BasePoiMail):
    # Override the plain '@property' with a simple property that we
    # can easily override in the tests.
    plain = u''


class TestNotifications(ptc.PoiTestCase):
    """Test email notifications

    For the moment this just checks if some plain text and html is
    being generated.
    """

    def _makeOne(self):
        # We may want to create a tracker and an issue and use the
        # issue as the context, but let's see how far we come with the
        # portal as context.
        return DummyPoiMail(self.portal, self.portal.REQUEST)

    def testCSS(self):
        # Check that the contents of the css file are being added to
        # the html by testing that at least some styling is ending up there.
        mailer = self._makeOne()
        self.assertTrue('font-size' in mailer.html)

    def testGoodRestructuredText(self):
        # Good restructured text gets interpreted:
        mailer = self._makeOne()
        mailer.plain = u"*bold* **italic**"
        self.assertTrue("<em>bold</em> <strong>italic</strong>" in mailer.html)

    def testBadRestructuredText(self):
        # Bad restructured text gets put in a 'pre' tag:
        mailer = self._makeOne()
        mailer.plain = u"This is **bad* restructured text"
        self.assertTrue("<pre>This is **bad* restructured text</pre>"
                        in mailer.html)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestNotifications))
    return suite



import unittest

import doctest

OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(module='Products.Poi.browser.response',
                             optionflags=OPTIONFLAGS),
        doctest.DocFileSuite('responses.txt',
                             package='Products.Poi.tests',
                             optionflags=OPTIONFLAGS),
        doctest.DocFileSuite('linkdetection.txt',
                             package='Products.Poi.tests',
                             optionflags=OPTIONFLAGS),
    ])



import logging

from AccessControl import Unauthorized
from Acquisition import aq_inner
from plone.namedfile import NamedBlobFile
from plone.namedfile.browser import Download as BlobDownload
from Products.Archetypes.atapi import DisplayList
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as PMF
from Products.CMFPlone.utils import safe_unicode
from Products.Five.browser import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from plone.memoize.view import memoize
from zope.cachedescriptors.property import Lazy
from zope.i18n import translate
from zope.interface import implements
from zope.lifecycleevent import modified

from Products.Poi import PoiMessageFactory as _
from Products.Poi import permissions
from Products.Poi.adapters import IResponseContainer
from Products.Poi.adapters import Response
from Products.Poi.browser.interfaces import IResponseAdder
from Products.Poi.config import DEFAULT_ISSUE_MIME_TYPE
from Products.Poi.utils import normalize_filename


logger = logging.getLogger('Poi')


def pretty_size(size):
    if size <= 0:
        return "0 Kb"
    kb = size / 1024
    size = "%d Kb" % kb
    if kb > 999:
        mb = kb / 1024
        size = "%d Mb" % mb
        if mb > 999:
            gb = mb / 1024
            size = "%d Gb" % gb
    return size


def voc2dict(vocab, current=None):
    """Make a dictionary from a vocabulary.

    >>> from Products.Archetypes.atapi import DisplayList
    >>> vocab = DisplayList()
    >>> vocab.add('a', "The letter A")
    >>> voc2dict(vocab)
    [{'checked': '', 'value': 'a', 'label': 'The letter A'}]
    >>> vocab.add('b', "The letter B")
    >>> voc2dict(vocab)
    [{'checked': '', 'value': 'a', 'label': 'The letter A'},
    {'checked': '', 'value': 'b', 'label': 'The letter B'}]
    >>> voc2dict(vocab, current='c')
    [{'checked': '', 'value': 'a', 'label': 'The letter A'},
    {'checked': '', 'value': 'b', 'label': 'The letter B'}]
    >>> voc2dict(vocab, current='b')
    [{'checked': '', 'value': 'a', 'label': 'The letter A'},
    {'checked': 'checked', 'value': 'b', 'label': 'The letter B'}]

    """
    options = []
    for value, label in vocab.items():
        checked = (value == current) and "checked" or ""
        options.append(dict(value=value, label=label,
                            checked=checked))
    return options


class Base(BrowserView):
    """Base view for PoiIssues.

    Mostly meant as helper for adding a response.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.folder = IResponseContainer(context)
        self.mimetype = DEFAULT_ISSUE_MIME_TYPE
        self.use_wysiwyg = (self.mimetype == 'text/html')

    def responses(self):
        context = aq_inner(self.context)
        trans = context.portal_transforms
        items = []
        linkDetection = context.linkDetection
        for id, response in enumerate(self.folder):
            if response is None:
                # Has been removed.
                continue
            # Use the already rendered response when available
            if response.rendered_text is None:
                rendering_success = True
                if response.mimetype in ('text/html', 'text/x-html-safe'):
                    html = response.text
                else:
                    html = trans.convertTo('text/html',
                                           response.text,
                                           mimetype=response.mimetype)
                    if html is None:
                        logger.debug("Conversion to text/html failed for "
                                     "response id %s of %s", id,
                                     context.absolute_url())
                        html = u''
                        rendering_success = False
                    else:
                        html = html.getData()
                if rendering_success:
                    # Detect links like #1 and r1234
                    html = linkDetection(html)
                    response.rendered_text = html

            html = response.rendered_text or u''
            info = dict(id=id,
                        response=response,
                        attachment=self.attachment_info(id),
                        html=html)
            items.append(info)
        return items

    @property
    @memoize
    def portal_url(self):
        context = aq_inner(self.context)
        plone = context.restrictedTraverse('@@plone_portal_state')
        return plone.portal_url()

    def attachment_info(self, id):
        """Get icon and other info for attachment

        Taken partly from Archetypes/skins/archetypes/getBestIcon.py
        """
        context = aq_inner(self.context)
        response = self.folder[id]
        attachment = response.attachment
        if attachment is None:
            return None

        from zExceptions import NotFound

        icon = None
        mtr = getToolByName(context, 'mimetypes_registry', None)
        if mtr is None:
            icon = context.getIcon()
        content_type = getattr(
            attachment, 'contentType', 'application/octet-stream')
        size = getattr(attachment, 'get_size', 0)
        if callable(size):
            size = size()
        lookup = mtr.lookup(content_type)
        if lookup:
            mti = lookup[0]
            try:
                context.restrictedTraverse(mti.icon_path)
                icon = mti.icon_path
            except (NotFound, KeyError, AttributeError):
                pass
        if icon is None:
            icon = context.getIcon()
        filename = getattr(attachment, 'filename', '')
        info = dict(
            icon=self.portal_url + '/' + icon,
            url=context.absolute_url() +
            '/@@poi_response_attachment?response_id=' + str(id),
            content_type=content_type,
            size=pretty_size(size),
            filename=filename,
        )
        return info

    @Lazy
    def memship(self):
        context = aq_inner(self.context)
        return getToolByName(context, 'portal_membership')

    @property
    @memoize
    def can_edit_response(self):
        context = aq_inner(self.context)
        return self.memship.checkPermission('Poi: Edit response', context)

    @property
    @memoize
    def can_delete_response(self):
        context = aq_inner(self.context)
        return self.memship.checkPermission('Delete objects', context)

    def validate_response_id(self):
        """Validate the response id from the request.

        Return -1 if for example the response id does not exist.
        Return the response id otherwise.

        Side effect: an informative status message is set.
        """
        status = IStatusMessage(self.request)
        response_id = self.request.form.get('response_id', None)
        if response_id is None:
            msg = _(u"No response selected.")
            msg = translate(msg, 'Poi', context=self.request)
            status.addStatusMessage(msg, type='error')
            return -1
        else:
            try:
                response_id = int(response_id)
            except ValueError:
                msg = _(u"Response id ${response_id} is no integer.",
                        mapping=dict(response_id=response_id))
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='error')
                return -1
            if response_id >= len(self.folder):
                msg = _(u"Response id ${response_id} does not exist.",
                        mapping=dict(response_id=response_id))
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='error')
                return -1
            else:
                return response_id
        # fallback
        return -1

    @property
    def severity(self):
        context = aq_inner(self.context)
        return context.getSeverity()

    @property
    def targetRelease(self):
        context = aq_inner(self.context)
        return context.getTargetRelease()

    @property
    def responsibleManager(self):
        context = aq_inner(self.context)
        return context.getResponsibleManager()

    @property
    @memoize
    def transitions_for_display(self):
        """Display the available transitions for this issue.
        """
        context = aq_inner(self.context)
        if not self.memship.checkPermission(permissions.ModifyIssueState,
                                            context):
            return []
        wftool = getToolByName(context, 'portal_workflow')
        transitions = []
        transitions.append(dict(value='', label=PMF(u'No change'),
                                checked="checked"))
        for tdef in wftool.getTransitionsFor(context):
            transitions.append(dict(value=tdef['id'],
                                    label=tdef['title_or_id'], checked=''))
        return transitions

    @property
    def available_transitions(self):
        """Get the available transitions for this issue.
        """
        return [x['value'] for x in self.transitions_for_display]

    @property
    def severities_for_display(self):
        """Get the available severities for this issue.
        """
        vocab = self.available_severities
        options = []
        for value in vocab:
            checked = (value == self.severity) and "checked" or ""
            options.append(dict(value=value, label=value,
                                checked=checked))
        return options

    @property
    @memoize
    def available_severities(self):
        """Get the available severities for this issue.
        """
        # get vocab from tracker so use aq_inner
        context = aq_inner(self.context)
        if not self.memship.checkPermission(
                permissions.ModifyIssueSeverity, context):
            return []
        return context.getAvailableSeverities()

    @property
    def releases_for_display(self):
        """Get the releases from the project.

        Usually nothing, unless you use Poi in combination with
        PloneSoftwareCenter.
        """
        vocab = self.available_releases
        current = self.targetRelease
        return voc2dict(vocab, current)

    @property
    @memoize
    def available_releases(self):
        """Get the releases from the project.

        Usually nothing, unless you use Poi in combination with
        PloneSoftwareCenter.
        """
        # get vocab from issue
        context = aq_inner(self.context)
        if not self.memship.checkPermission(
                permissions.ModifyIssueTargetRelease, context):
            return DisplayList()
        return context.getReleasesVocab()

    @property
    def show_target_releases(self):
        """Should the option for selecting a target release be shown?

        There is always at least one option: None.  So only show when
        there is more than one option.
        """
        return len(self.available_releases) > 1

    @property
    def managers_for_display(self):
        """Get the tracker managers.
        """
        vocab = self.available_managers
        return voc2dict(vocab, self.responsibleManager)

    @property
    @memoize
    def available_managers(self):
        """Get the tracker managers.
        """
        # get vocab from issue
        context = aq_inner(self.context)
        if not self.memship.checkPermission(
                permissions.ModifyIssueAssignment, context):
            return DisplayList()
        return context.getManagersVocab()

    @property
    @memoize
    def upload_allowed(self):
        """Is the user allowed to upload on attachment?
        """
        context = aq_inner(self.context)
        return self.memship.checkPermission(
            permissions.UploadAttachment, context)


class AddForm(Base):
    implements(IResponseAdder)

    def __init__(self, context, request, view):
        super(AddForm, self).__init__(context, request)
        self.__parent__ = view

    def __getitem__(self, name):
        # If we do not add this method, you keep getting a warning at Zope
        # startup: Init Class Products.Five.viewlet.manager.<ViewletManager
        # providing IResponseAdder> has a security declaration for nonexistent
        # method '__getitem__'.
        raise IndexError(
            'This is not meant to be used as viewlet manager. '
            'It is just a contentprovider.')

    def update(self):
        pass

    def render(self):
        # self.template is defined in zcml
        return self.template()


class Create(Base):

    def determine_response_type(self, response):
        """Return a string indicating the type of response this is.
        """
        responseCreator = response.creator
        if responseCreator == '(anonymous)':
            return 'additional'

        issue = aq_inner(self.context)
        if responseCreator == issue.Creator():
            return 'clarification'

        if responseCreator in self.available_managers:
            return 'reply'

        # default:
        return 'additional'

    def __call__(self):
        form = self.request.form
        context = aq_inner(self.context)
        if not self.memship.checkPermission('Poi: Add Response', context):
            raise Unauthorized

        response_text = form.get('response', u'')
        new_response = Response(response_text)
        new_response.mimetype = self.mimetype
        new_response.type = self.determine_response_type(new_response)

        issue_has_changed = False
        transition = form.get('transition', u'')
        if transition and transition in self.available_transitions:
            wftool = getToolByName(context, 'portal_workflow')
            before = wftool.getInfoFor(context, 'review_state')
            before = wftool.getTitleForStateOnType(before, 'PoiIssue')
            wftool.doActionFor(context, transition)
            after = wftool.getInfoFor(context, 'review_state')
            after = wftool.getTitleForStateOnType(after, 'PoiIssue')
            new_response.add_change('review_state', _(u'Issue state'),
                                    before, after)
            issue_has_changed = True

        options = [
            ('severity', _(u'Severity'), 'available_severities'),
            ('responsibleManager', _(u'Responsible manager'),
             'available_managers'),
        ]
        # Changes that need to be applied to the issue (apart from
        # workflow changes that need to be handled separately).
        changes = {}
        for option, title, vocab in options:
            new = form.get(option, u'')
            if new and new in self.__getattribute__(vocab):
                current = self.__getattribute__(option)
                if current != new:
                    changes[option] = new
                    new_response.add_change(option, title,
                                            current, new)
                    issue_has_changed = True

        new = form.get('targetRelease', u'')
        if new and new in self.available_releases:
            current = self.targetRelease
            if current != new:
                # from value (uid) to key (id)
                new_label = self.available_releases.getValue(new)
                current_label = self.available_releases.getValue(current)
                changes['targetRelease'] = new
                new_response.add_change('targetRelease', _(u'Target release'),
                                        current_label, new_label)
                issue_has_changed = True

        attachment = form.get('attachment')
        if attachment:
            # File(id, title, file)
            # data = File(attachment.filename, attachment.filename, attachment)
            data = NamedBlobFile(
                attachment, filename=safe_unicode(attachment.filename))
            new_response.attachment = data
            issue_has_changed = True

        if len(response_text) == 0 and not issue_has_changed:
            status = IStatusMessage(self.request)
            msg = _(u"No response text added and no issue changes made.")
            msg = translate(msg, 'Poi', context=self.request)
            status.addStatusMessage(msg, type='error')
        else:
            # Apply changes to issue
            context.update(**changes)
            # Add response
            self.folder.add(new_response)
        self.request.response.redirect(context.absolute_url())


class Edit(Base):

    @property
    @memoize
    def response(self):
        form = self.request.form
        response_id = form.get('response_id', None)
        if response_id is None:
            return None
        try:
            response_id = int(response_id)
        except ValueError:
            return None
        if response_id >= len(self.folder):
            return None
        return self.folder[response_id]

    @property
    def response_found(self):
        return self.response is not None


class Save(Base):

    def __call__(self):
        form = self.request.form
        context = aq_inner(self.context)
        status = IStatusMessage(self.request)
        if not self.can_edit_response:
            msg = _(u"You are not allowed to edit responses.")
            msg = translate(msg, 'Poi', context=self.request)
            status.addStatusMessage(msg, type='error')
        else:
            response_id = form.get('response_id', None)
            if response_id is None:
                msg = _(u"No response selected for saving.")
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='error')
            elif self.folder[response_id] is None:
                msg = _(u"Response does not exist anymore; perhaps it was "
                        "removed by another user.")
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='error')
            else:
                response = self.folder[response_id]
                response_text = form.get('response', u'')
                response.text = response_text
                # Remove cached rendered response.
                response.rendered_text = None
                msg = _(u"Changes saved to response id ${response_id}.",
                        mapping=dict(response_id=response_id))
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='info')
                # Fire event.  We put the context in the descriptions
                # so event handlers can use this fully acquisition
                # wrapped object to do their thing.  Feels like
                # cheating, but it gets the job done.  Arguably we
                # could turn the two arguments around and signal that
                # the issue has changed, with the response in the
                # event descriptions.
                modified(response, context)
        self.request.response.redirect(context.absolute_url())


class Delete(Base):

    def __call__(self):
        context = aq_inner(self.context)
        status = IStatusMessage(self.request)

        if not self.can_delete_response:
            msg = _(u"You are not allowed to delete responses.")
            msg = translate(msg, 'Poi', context=self.request)
            status.addStatusMessage(msg, type='error')
        else:
            response_id = self.request.form.get('response_id', None)
            if response_id is None:
                msg = _(u"No response selected for removal.")
                msg = translate(msg, 'Poi', context=self.request)
                status.addStatusMessage(msg, type='error')
            else:
                try:
                    response_id = int(response_id)
                except ValueError:
                    msg = _(u"Response id ${response_id} is no integer so it "
                            "cannot be removed.",
                            mapping=dict(response_id=response_id))
                    msg = translate(msg, 'Poi', context=self.request)
                    status.addStatusMessage(msg, type='error')
                    self.request.response.redirect(context.absolute_url())
                    return
                if response_id >= len(self.folder):
                    msg = _(u"Response id ${response_id} does not exist so it "
                            "cannot be removed.",
                            mapping=dict(response_id=response_id))
                    msg = translate(msg, 'Poi', context=self.request)
                    status.addStatusMessage(msg, type='error')
                else:
                    self.folder.delete(response_id)
                    msg = _(u"Removed response id ${response_id}.",
                            mapping=dict(response_id=response_id))
                    msg = translate(msg, 'Poi', context=self.request)
                    status.addStatusMessage(msg, type='info')
        self.request.response.redirect(context.absolute_url())


class Download(Base, BlobDownload):
    """Download the attachment of a response.
    """

    def __call__(self):
        context = aq_inner(self.context)
        request = self.request
        response_id = self.validate_response_id()
        file = None
        if response_id != -1:
            response = self.folder[response_id]
            file = response.attachment
            if file is None:
                status = IStatusMessage(request)
                msg = _(u"Response id ${response_id} has no attachment.",
                        mapping=dict(response_id=response_id))
                msg = translate(msg, 'Poi', context=context)
                status.addStatusMessage(msg, type='error')
        if file is None:
            request.response.redirect(context.absolute_url())

        # From now on we know that an attachment exists.
        # Set attributes on self, so BlobDownload can do its work.
        self.fieldname = 'attachment'
        self.file = file
        filename = getattr(file, 'filename', self.fieldname)
        if filename is not None:
            filename = normalize_filename(filename, request)
        self.filename = filename
        return BlobDownload.__call__(self)

    def _getFile(self):
        return self.file



# -*- coding: utf-8 -*-

PROJECTNAME = "Poi"

# Permissions
DEFAULT_ADD_CONTENT_PERMISSION = "Add portal content"
ADD_CONTENT_PERMISSIONS = {
    'PoiTracker': 'Poi: Add Tracker',
    'PoiIssue': 'Poi: Add Issue',
}

product_globals = globals()

DESCRIPTION_LENGTH = 200
PSC_TRACKER_ID = 'issues'

# Add text/html to the list of mimetypes to allow HTML/kupu
# issue/response text.
ISSUE_MIME_TYPES = ('text/x-web-intelligent', 'text/plain')
DEFAULT_ISSUE_MIME_TYPE = 'text/x-web-intelligent'
# Patterns used for recognizing links to issues and revisions:
ISSUE_RECOGNITION_PATTERNS = \
    [r'\B#[1-9][0-9]*\b', r'\bissue:[1-9][0-9]*\b',
     r'\bticket:[1-9][0-9]*\b', r'\bbug:[1-9][0-9]*\b']
REVISION_RECOGNITION_PATTERNS = \
    [r'\br[0-9]+\b', r'\bchangeset:[0-9]+\b', r'\B\[[0-9]+\]\B']
# Template to use when recognizing a link to another issue:
ISSUE_LINK_TEMPLATE = '<a href="%(base_url)s/%(bug)s">%(linktext)s</a>'



import unittest

import doctest
from Testing.ZopeTestCase import FunctionalDocFileSuite as Suite

from Products.Poi.tests.ptc import PoiFunctionalTestCase

OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


def test_suite():
    suites = [
        Suite('browser.txt',
              package='Products.Poi.tests',
              optionflags=OPTIONFLAGS,
              test_class=PoiFunctionalTestCase),
    ]
    return unittest.TestSuite(suites)



# -*- coding: utf-8 -*-
from Testing import ZopeTestCase
from collective.watcherlist.interfaces import IWatcherList

from Products.Poi.events import sendResponseNotificationMail
from Products.Poi.tests import ptc

default_user = ZopeTestCase.user_name


class TestTracker(ptc.PoiTestCase):
    """Test tracker functionality"""

    def afterSetUp(self):
        self.tracker = self.createTracker(self.folder, 'issue-tracker')
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member2', 'Member Two', 'member2@example.com',
                       ['Member'], '2005-01-01')

    def testEditTracker(self):
        self.tracker.setTitle('title')
        self.tracker.setDescription('description')
        self.tracker.setHelpText('help text')
        self.tracker.setAvailableAreas(
            ({'id': 'area', 'title': 'Area', 'description': 'Issue area'}, ))
        self.tracker.setAvailableIssueTypes(
            ({'id': 'type', 'title': 'Type', 'description': 'Issue type'}, ))
        self.tracker.setAvailableSeverities(('one', 'two'))
        self.tracker.setDefaultSeverity('two')
        self.tracker.setAvailableReleases(('1.0', '2.0'))
        self.tracker.setManagers(('member1', 'member2'))
        self.tracker.setSendNotificationEmails(False)
        self.tracker.setMailingList('list@example.com')

        self.assertEqual(self.tracker.Title(), 'title')
        self.assertEqual(self.tracker.Description(), 'description')
        self.assertEqual(self.tracker.getHelpText(), '<p>help text</p>')
        self.assertEqual(
            self.tracker.getAvailableAreas(),
            ({'id': 'area', 'title': 'Area', 'description': 'Issue area'}, ))
        self.assertEqual(
            self.tracker.getAvailableIssueTypes(),
            ({'id': 'type', 'title': 'Type', 'description': 'Issue type'}, ))
        self.assertEqual(self.tracker.getAvailableSeverities(), ('one', 'two'))
        self.assertEqual(self.tracker.getDefaultSeverity(), 'two')
        self.assertEqual(self.tracker.getAvailableReleases(), ('1.0', '2.0'))
        self.assertEqual(self.tracker.getManagers(), ('member1', 'member2'))
        self.assertEqual(self.tracker.getSendNotificationEmails(), False)
        self.assertEqual(self.tracker.getMailingList(), 'list@example.com')

    def testDataGridFields(self):
        """
        The DataGridFields should have at least one entry, as they are
        required.  We get problems when adding an Issue if we are not
        careful.  See http://plone.org/products/poi/issues/139

        """
        # When adding/editing through the web we always have a hidden
        # entry:
        hidden_entry = {'description': '',
                        'id': '', 'orderindex_':
                        'template_row_marker', 'title': ''}
        # This is what a real entry looks like:
        real_entry = {'description': 'Something nice.',
                      'id': 'something',
                      'orderindex_': '1',
                      'title': 'Something'}

        # Test the availableAreas field.
        field = self.tracker.getField('availableAreas')
        input = [hidden_entry]
        errors = {}
        self.assertEqual(
            field.validate(input, self.tracker, errors),
            u'Areas is required, please correct.')
        self.assertTrue('availableAreas' in errors)
        self.assertEqual(errors['availableAreas'],
                         'Areas is required, please correct.')
        input = [real_entry, hidden_entry]
        errors = {}
        self.assertEqual(field.validate(input, self.tracker, errors), None)
        self.assertFalse(errors)

        # Test the availableIssueTypes field.
        field = self.tracker.getField('availableIssueTypes')
        input = [hidden_entry]
        errors = {}
        self.assertEqual(
            field.validate(input, self.tracker, errors),
            u'Issue types is required, please correct.')
        self.assertTrue('availableIssueTypes' in errors)
        self.assertEqual(errors['availableIssueTypes'],
                         'Issue types is required, please correct.')
        input = [real_entry, hidden_entry]
        errors = {}
        self.assertEqual(field.validate(input, self.tracker, errors), None)
        self.assertFalse(errors)

    def testValidateTrackerManagers(self):
        self.failUnless(self.tracker.validate_managers(('member1', )) is None)
        self.failIf(self.tracker.validate_managers(('memberX', )) is None)
        self.failIf(
            self.tracker.validate_managers(('member1', 'memberX')) is None)

    def testManagersGetLocalRole(self):
        roles = self.tracker.get_local_roles_for_userid
        self.failIf('Manager' in roles('member1'))
        self.failIf('TrackerManager' in roles('member1'))
        self.tracker.setManagers(('member1', ))
        self.failIf('Manager' in roles('member1'))
        self.failUnless('TrackerManager' in roles('member1'))
        self.tracker.setManagers(('member2', ))
        self.failIf('TrackerManager' in roles('member1'))
        self.failUnless('TrackerManager' in roles('member2'))
        # Now we mess with local roles and see if setting the tracker
        # managers fixes it.
        self.tracker.manage_delLocalRoles(['member2'])
        self.tracker.setManagers(('member2', ))
        self.failUnless('TrackerManager' in roles('member2'))
        self.tracker.manage_setLocalRoles(
            'member2', ['Owner', 'TrackerManager'])
        self.tracker.manage_setLocalRoles('member1', ['Reviewer'])
        self.tracker.setManagers(('member1', 'member2'))
        self.failUnless('TrackerManager' in roles('member2'))
        self.failUnless('TrackerManager' in roles('member1'))
        self.failUnless('Owner' in roles('member2'))
        self.failUnless('Reviewer' in roles('member1'))
        self.tracker.setManagers(('member2', ))
        self.failUnless('Reviewer' in roles('member1'))

    def testUpgradeManagers(self):
        roles = self.tracker.get_local_roles_for_userid
        self.tracker.setManagers(('member1', 'member2', ))
        # Mess things up
        self.tracker.manage_setLocalRoles(
            'member1', ['Owner', 'Manager', 'TrackerManager'])
        self.tracker.manage_setLocalRoles('member2', ['Reviewer'])
        # Our upgrade step should fix it.
        from Products.Poi.migration import update_tracker_managers
        update_tracker_managers(self.portal, testing=True)

        self.failUnless('Owner' in roles('member1'))
        self.failUnless('TrackerManager' in roles('member1'))
        self.failIf('Manager' in roles('member1'))

        self.failUnless('Reviewer' in roles('member2'))
        self.failUnless('TrackerManager' in roles('member2'))
        self.failIf('Manager' in roles('member2'))

    def testIsUsingReleases(self):
        self.tracker.setAvailableReleases(())
        self.failIf(self.tracker.isUsingReleases())
        self.tracker.setAvailableReleases(('1.0', '2.0'))
        self.failUnless(self.tracker.isUsingReleases())


class TestEmailNotifications(ptc.PoiTestCase):
    """Test getting email addresses and sending email notifications"""

    def afterSetUp(self):
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member2', 'Member Two', 'member2@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member3', 'Member Three', 'member3@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(
            self.folder, 'issue-tracker', managers=('member1', 'member2'),
            sendNotificationEmails=True)

    def testGetAddressesWithNotificationsOff(self):
        self.tracker.setSendNotificationEmails(False)
        issue = self.createIssue(
            self.tracker, contactEmail='submitter@example.com',
            watchers=('member2', 'member3'))
        watcherlist = IWatcherList(issue)
        # We have two watchers directly on this issue, plus the submitter:
        watchers = watcherlist.watchers
        self.assertEqual(len(watchers), 3)
        # But since emails are not sent, we have zero addresses:
        addresses = watcherlist.addresses
        self.assertEqual(len(addresses), 0)

    def testGetAddressesOnNewIssue(self):
        addresses = IWatcherList(self.tracker).addresses
        self.assertEqual(len(addresses), 2)
        self.failUnless('member1@example.com' in addresses)
        self.failUnless('member2@example.com' in addresses)

    def testGetAddressesOnNewIssueWithList(self):
        self.tracker.setMailingList('list@example.com')
        addresses = IWatcherList(self.tracker).addresses
        # Addresses are the mailing list and the tracker managers.
        self.assertEqual(len(addresses), 3)
        self.failUnless('list@example.com' in addresses)

    def testGetAddressesOnNewResponse(self):
        issue = self.createIssue(
            self.tracker, contactEmail='submitter@example.com',
            watchers=('member2', 'member3'))
        addresses = IWatcherList(issue).addresses
        self.assertEqual(len(addresses), 4)
        self.failUnless('member1@example.com' in addresses)
        self.failUnless('member2@example.com' in addresses)
        self.failUnless('member3@example.com' in addresses)
        self.failUnless('submitter@example.com' in addresses)
        # A mail is sent immediately on creation of this issue.
        self.assertEqual(len(self.portal.MailHost.messages), 4)

    def testGetAddressesOnNewResponseWithList(self):
        self.tracker.setMailingList('list@example.com')
        issue = self.createIssue(
            self.tracker, contactEmail='submitter@example.com',
            watchers=('member2', 'member3'))
        addresses = IWatcherList(issue).addresses
        self.assertEqual(len(addresses), 5)
        # mailing list:
        self.failUnless('list@example.com' in addresses)
        # submitter:
        self.failUnless('submitter@example.com' in addresses)
        # tracker manager:
        self.failUnless('member1@example.com' in addresses)
        # direct subscribers:
        self.failUnless('member2@example.com' in addresses)
        self.failUnless('member3@example.com' in addresses)
        # A mail is sent immediately on creation of this issue.
        self.assertEqual(len(self.portal.MailHost.messages), 5)

    def testGetTagsInUse(self):
        self.createIssue(self.tracker, tags=('A', 'B'))
        self.createIssue(self.tracker, tags=('B', 'C'))
        self.createIssue(self.tracker, tags=('A', 'D'))
        self.assertEqual(self.tracker.getTagsInUse(), ['A', 'B', 'C', 'D'])

    # The following tests don't map directly to functional methods but are
    # meant to make sure no errors arise from sending emails
    # -- begin email tests

    def testNewIssueEmail(self):
        self.tracker.setSendNotificationEmails(True)
        self.tracker.update(title='Random Tracker')
        # Just creating it should be enough to send an email.
        self.createIssue(self.tracker,
                         contactEmail='submitter@example.com',
                         watchers=('member1', 'member2'))
        # A mail is sent immediately on creation of this issue.
        self.assertEqual(len(self.portal.MailHost.messages), 3)

    def testSpecialCharacterIssueEmail(self):
        self.tracker.setSendNotificationEmails(True)
        self.tracker.update(title='Random Tracker')
        issue = self.createIssue(
            self.tracker,
            title="accented vocals: à è ì",
            contactEmail='submitter@example.com',
            watchers=('member1', 'member2'))
        self.createResponse(
            issue, text="more accented vocals: ò ù")
        sendResponseNotificationMail(issue)

        # Now try a different charset
        self.portal.email_charset = 'iso-8859-1'
        issue = self.createIssue(
            self.tracker,
            title=u"accented vocals: à è ì ò ù".encode('utf-8'),
            contactEmail='submitter@example.com',
            watchers=('member1', 'member2'))
        self.createResponse(
            issue, text=u"more accented vocals: ò ù".encode('iso-8859-1'))
        sendResponseNotificationMail(issue)

    def testNewResponseEmail(self):
        self.tracker.setSendNotificationEmails(True)
        self.tracker.update(title='Random Tracker')
        issue = self.createIssue(self.tracker,
                                 contactEmail='submitter@example.com',
                                 watchers=('member1', 'member2'))
        self.createResponse(issue)
        sendResponseNotificationMail(issue)

    def testResolvedEmail(self):
        self.tracker.setSendNotificationEmails(True)
        self.tracker.update(title='Random Tracker')

        issue = self.createIssue(self.tracker,
                                 contactEmail='submitter@example.com',
                                 watchers=('member1', 'member2'))
        self.loginAsPortalOwner()
        workflow = self.portal.portal_workflow
        workflow.doActionFor(issue, 'resolve-unconfirmed')

    # -- end email tests


class TestTrackerSearch(ptc.PoiTestCase):
    """Test tracker search functionality"""

    def afterSetUp(self):
        self.tracker = self.createTracker(self.folder, 'issue-tracker')
        self.workflow = self.portal.portal_workflow
        self.issuefolder = self.tracker.restrictedTraverse('@@issuefolder')

    def testGetFilteredIssuesById(self):
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        issues = [b.getId for b in self.issuefolder.getFilteredIssues(id='1')]
        self.assertEqual(issues, ['1'])
        issues = [b.getId for b in self.issuefolder.getFilteredIssues(id='2')]
        self.assertEqual(issues, ['2'])

    def testGetFilteredIssesByRelease(self):
        self.createIssue(self.tracker, release='2.0')
        self.createIssue(self.tracker, release='2.0')
        self.createIssue(self.tracker, release='1.0')
        issues = sorted([b.getId
                         for b in self.issuefolder.getFilteredIssues(release='2.0')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId
                  for b in self.issuefolder.getFilteredIssues(release='1.0')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByArea(self):
        self.createIssue(self.tracker, area='ui')
        self.createIssue(self.tracker, area='ui')
        self.createIssue(self.tracker, area='functionality')
        issues = sorted([b.getId
                         for b in self.issuefolder.getFilteredIssues(area='ui')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(area='functionality')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByIssueType(self):
        self.createIssue(self.tracker, issueType='bug')
        self.createIssue(self.tracker, issueType='bug')
        self.createIssue(self.tracker, issueType='feature')
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(issueType='bug')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(issueType='feature')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesBySeverity(self):
        self.createIssue(self.tracker, severity='Medium')
        self.createIssue(self.tracker, severity='Medium')
        self.createIssue(self.tracker, severity='Critical')
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(severity='Medium')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(severity='Critical')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByTargetRelease(self):
        self.createIssue(self.tracker, targetRelease='2.0')
        self.createIssue(self.tracker, targetRelease='2.0')
        self.createIssue(self.tracker, targetRelease='1.0')
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(targetRelease='2.0')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(targetRelease='1.0')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByState(self):
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        self.setRoles(['TrackerManager'])
        self.workflow.doActionFor(self.tracker['3'], 'accept-unconfirmed')
        self.setRoles(['Member'])
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(state='unconfirmed')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(state='open')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByCreator(self):
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)
        self.tracker['1'].setCreators(('some_member', ))
        self.tracker['1'].reindexObject()
        self.tracker['2'].setCreators(('some_member', ))
        self.tracker['2'].reindexObject()
        self.tracker['3'].setCreators(('another_member', ))
        self.tracker['3'].reindexObject()

        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(creator='some_member')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(creator='another_member')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByResponsible(self):
        self.createIssue(self.tracker, responsibleManager='manager1')
        self.createIssue(self.tracker, responsibleManager='manager1')
        self.createIssue(self.tracker, responsibleManager='manager2')
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(responsible='manager1')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(responsible='manager2')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByTags(self):
        self.createIssue(self.tracker, tags=('A', 'B'))
        self.createIssue(self.tracker, tags=('B', 'C'))
        self.createIssue(self.tracker, tags=('A', 'D'))
        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(tags='B')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=('A', 'D'))]
        issues.sort()
        self.assertEqual(issues, ['1', '3'])
        # Operator is 'or' by default, so this should give the same results.
        issues = [b.getId for b in self.issuefolder.getFilteredIssues(
            tags=dict(query=('A', 'D'), operator='or'))]
        issues.sort()
        self.assertEqual(issues, ['1', '3'])
        # Now try with 'and'
        issues = [b.getId for b in self.issuefolder.getFilteredIssues(
            tags=dict(query=('A', 'D'), operator='and'))]
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesByIssueText(self):
        self.createIssue(self.tracker, details="foo")
        issues = sorted([b.getId
                         for b in self.issuefolder.getFilteredIssues(text='foo')])
        self.assertEqual(issues, ['1'])
        issues = [b.getId
                  for b in self.issuefolder.getFilteredIssues(text='bar')]
        self.assertEqual(len(issues), 0)

    def testGetFilteredIssesByResponseText(self):
        self.createIssue(self.tracker, details="foo")
        self.createIssue(self.tracker)
        self.createIssue(self.tracker)

        self.createResponse(self.tracker['2'], text='foo')
        self.createResponse(self.tracker['3'], text='bar')

        issues = sorted([b.getId
                         for b in self.issuefolder.getFilteredIssues(text='foo')])
        self.assertEqual(issues, ['1', '2'])
        issues = [b.getId
                  for b in self.issuefolder.getFilteredIssues(text='bar')]
        issues.sort()
        self.assertEqual(issues, ['3'])

    def testGetFilteredIssesComplex(self):
        self.createIssue(
            self.tracker, details="foo", area="ui", issueType='feature')
        self.createIssue(self.tracker, area="ui", issueType="bug")
        self.createIssue(
            self.tracker, area="functionality", details="foo", issueType='bug')

        issues = sorted([b.getId for b in
                         self.issuefolder.getFilteredIssues(text='foo', area='ui')])
        self.assertEqual(issues, ['1'])

        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(area='ui',
                                                     issueType='feature')]
        issues.sort()
        self.assertEqual(issues, ['1'])

        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(area='ui',
                                                     issueType=['feature', 'bug'])]
        issues.sort()
        self.assertEqual(issues, ['1', '2'])

    def testSubjectTolerance(self):
        self.createIssue(self.tracker, details="foo", area="ui",
                         issueType='feature', tags=('A'))
        issues = [
            b.getId for b in
            self.issuefolder.getFilteredIssues(tags=dict(operator='and'))]
        self.assertEqual(issues, ['1'])

        issues = [
            b.getId for b in
            self.issuefolder.getFilteredIssues(Subject=dict(operator='or'))]
        self.assertEqual(issues, ['1'])

        class FakeQuery:
            """Fake query for use in the poi_issue_search_form.

            When filling in the poi_issue_search_form you do not get a
            dict, but an InstanceType, so an old style class.  That
            can give problems, so we fake it here.
            """

            def __init__(self, operator=None, query=None):
                if operator is not None:
                    self.operator = operator
                if query is not None:
                    self.query = query

            def __getitem__(self, key):
                return self.__dict__[key]

        # We repeat the previous two tests but now with FakeOperators
        # instead of dicts.
        tags = FakeQuery(operator='and')
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, ['1'])
        tags = FakeQuery(operator='or')
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, ['1'])

        # Might as well throw in a few more tests, as we do not yet
        # catch all errors.

        tags = FakeQuery(query=['A'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, ['1'])
        tags = FakeQuery(operator='and', query=['A'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, ['1'])
        tags = FakeQuery(operator='and', query=['A', 'B'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, [])
        tags = FakeQuery(operator='or', query=['A', 'B'])
        issues = [b.getId for b in
                  self.issuefolder.getFilteredIssues(tags=tags)]
        self.assertEqual(issues, ['1'])


class TestLinkDetection(ptc.PoiTestCase):
    """Test link detection functionality"""

    def afterSetUp(self):
        self.tracker = self.createTracker(self.folder, 'issue-tracker')

    def testLinksInIssues(self):
        """These are more tests for issues really,
        but they also test the tracker indirectly.
        """

        # Create an issue.
        self.createIssue(self.tracker)

        # Link to that existing issue.
        issue = self.createIssue(self.tracker, details="#1")
        self.assertEqual(issue.getTaggedDetails(),
                         '<p><a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a></p>')

        # Issue #3 does not exist.
        issue.update(details="#3")
        self.assertEqual(self.tracker['2'].getTaggedDetails(),
                         '<p>#3</p>')

        # Link to an existing issue in the steps
        issue.update(steps="#1")
        self.assertEqual(self.tracker['2'].getTaggedSteps(),
                         '<p><a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a></p>')

    def testLinksToIssues(self):
        tracker = self.tracker

        # Text without anything special will simply be returned
        # unchanged:
        self.assertEqual(
            tracker.linkDetection("We are the knights who say 'Ni'!"),
            "We are the knights who say 'Ni'!")

        # Unicode should not give problems:
        self.assertEqual(
            tracker.linkDetection(u'\xfanicode'),
            u'\xfanicode')

        # We can ask this tracker to detect issues.  But it does
        # nothing with non existing issues:
        self.assertEqual(tracker.linkDetection("#1"), '#1')

        # Now we add an issue.  The link detection code searches for
        # issues in the portal catalog.  So we add issues there:
        self.createIssue(self.tracker, title="1")
        self.createIssue(self.tracker, title="2")

        # Now we should get html back when we ask for an issue number:
        self.assertEqual(
            tracker.linkDetection("#1"),
            '<a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a>')
        self.assertEqual(
            tracker.linkDetection("Links to #1 and #2."),
            'Links to <a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a> and <a href="http://nohost/plone/Members/test_user_1_/issue-tracker/2">#2</a>.')

        # We are not fooled by a non existing issue:
        self.assertEqual(
            tracker.linkDetection("Issue #1 and non-issue #3."),
            'Issue <a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a> and non-issue #3.')

        # Issues that are added to a different tracker only show up
        # for that tracker:
        tracker2 = self.createTracker(self.folder, 'tracker2')
        self.assertEqual(
            tracker2.linkDetection("#1"),
            '#1')
        self.createIssue(tracker2, title="1")
        self.assertEqual(
            tracker2.linkDetection("#1"),
            '<a href="http://nohost/plone/Members/test_user_1_/tracker2/1">#1</a>')

        # A combination of unicode and a link number should be possible::
        self.assertEqual(
            tracker.linkDetection(u'\xfanicode text with a link to #1'),
            u'\xfanicode text with a link to <a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a>')

    def testLinksToRevisions(self):
        tracker = self.tracker

        # We can link to revisions or changesets.  By default nothing
        # happens:

        self.assertEqual(tracker.linkDetection('r42'), 'r42')

        # We need to specify in the tracker where those links should
        # point to.  We could point to something silly:

        tracker.update(svnUrl="silly")
        self.assertEqual(
            tracker.linkDetection('r42'),
            '<a href="silly">r42</a>')

        # This is not very useful, as this is not really a link
        # (unless this is a relative link to some content with the id
        # 'silly') and it does nothing with the revision number.  The
        # *real* idea here is to specify a string with "%(rev)s" in
        # it.  At that point the revision number will be filled in.

        # You could point to revisions, for example the collective
        # Trac for Poi:

        tracker.update(
            svnUrl="http://dev.plone.org/collective/browser/Poi?%(rev)s")
        self.assertEqual(
            tracker.linkDetection('r42'),
            '<a href="http://dev.plone.org/collective/browser/Poi?42">r42</a>')

        # I myself like to point to the changesets:

        tracker.update(
            svnUrl="http://dev.plone.org/changeset/%(rev)s/collective")
        self.assertEqual(
            tracker.linkDetection('r42'),
            '<a href="http://dev.plone.org/changeset/42/collective">r42</a>')

        # Of course it is fine to combine issues and revisions:
        self.createIssue(tracker, title="1")
        self.assertEqual(
            tracker.linkDetection('Issue #1 is fixed in r42.'),
            'Issue <a href="http://nohost/plone/Members/test_user_1_/issue-tracker/1">#1</a> is fixed in <'
            'a href="http://dev.plone.org/changeset/42/collective">r42</a>.')


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestTracker))
    suite.addTest(makeSuite(TestTrackerSearch))
    suite.addTest(makeSuite(TestEmailNotifications))
    suite.addTest(makeSuite(TestLinkDetection))
    return suite



from Products.CMFCore.utils import getToolByName


def add_catalog_indexes(site, logger):
    """Add our indexes to the catalog.

    Doing it here instead of in profiles/default/catalog.xml means we
    do not need to reindex those indexes after every reinstall.
    """
    catalog = getToolByName(site, 'portal_catalog')
    indexes = catalog.indexes()
    wanted = ("getRelease", "getArea", "getIssueType", "getSeverity",
              "getTargetRelease", "getResponsibleManager")

    missing = [w for w in wanted if w not in indexes]
    if missing:
        for name in missing:
            catalog.addIndex(name, 'FieldIndex')
            logger.info("Added FieldIndex for field %s.", name)
        logger.info("Indexing new indexes %s.", ', '.join(missing))
        catalog.manage_reindexIndex(ids=missing)


def allow_psc_tracker(site, logger):
    """Add PoiPscTracker to allowed types in PSCProject.

    Fails gracefully if PSCProject does not exist.
    """
    types_tool = getToolByName(site, 'portal_types')
    fti = getattr(types_tool, 'PSCProject', None)
    if fti is None:
        logger.info("Could not add PoiPscTracker to allowed content types of "
                    "PSCProject as it does not exist.")
        return
    allowed_types = list(fti.allowed_content_types)
    if 'PoiPscTracker' in allowed_types:
        logger.info("PoiPscTracker is already in allowed content types of "
                    "PSCProject.")
    allowed_types.append('PoiPscTracker')
    fti.allowed_content_types = tuple(allowed_types)
    logger.info("Added PoiPscTracker to allowed content types PSCProject.")


def import_various(context):
    # Only run step if a flag file is present
    if context.readDataFile('poi_various.txt') is None:
        return
    logger = context.getLogger('Poi')
    site = context.getSite()
    add_catalog_indexes(site, logger)
    allow_psc_tracker(site, logger)



from AccessControl import Unauthorized
from AccessControl import getSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from Testing import ZopeTestCase

from Products.Poi.tests import ptc

ZopeTestCase.installProduct('PloneSoftwareCenter')


class TestPscTracker(ptc.PoiTestCase):
    """Test psc tracker functionality"""

    def reinstallProduct(self, name):
        """Re-installs a product into the Plone site."""
        sm = getSecurityManager()
        self.loginAsPortalOwner()
        try:
            qi = self.portal.portal_quickinstaller
            if qi.isProductInstalled(name):
                qi.reinstallProducts(name)
                self._refreshSkinData()
        finally:
            setSecurityManager(sm)

    def afterSetUp(self):
        self.addProduct('PloneSoftwareCenter')
        self.reinstallProduct('Poi')

        self.setRoles(['Manager'])
        self.portal.invokeFactory('PloneSoftwareCenter', 'psc')
        self.psc = self.portal.psc
        self.psc.invokeFactory('PSCProject', 'project')
        self.project = self.psc.project

    def testTrackedAllowedInProject(self):
        try:
            self.project.invokeFactory('PoiPscTracker', 'issues')
        except Unauthorized:
            self.fail()

    def testNoReleasesInSchema(self):
        self.project.invokeFactory('PoiPscTracker', 'issues')
        self.failIf('releases' in self.project.issues.Schema())

    def testReleasesVocab(self):
        self.project.invokeFactory('PoiPscTracker', 'issues')
        self.project.releases.invokeFactory('PSCRelease', '1.0')
        self.project.releases.invokeFactory('PSCRelease', '2.0')
        self.failUnless(self.project.issues.isUsingReleases())
        releasesVocab = self.project.issues.getReleasesVocab()
        self.assertEqual(len(releasesVocab), 2)
        self.assertEqual(releasesVocab.values(), ['1.0', '2.0'])

    def testTitle(self):
        self.project.invokeFactory('PoiPscTracker', 'issues')
        self.assertEqual(self.project.issues.Title(), "Issue tracker")

    def testExternalTitle(self):
        self.project.invokeFactory('PoiPscTracker', 'issues')
        self.project.setTitle("My project")
        self.assertEqual(self.project.issues.getExternalTitle(),
                         "My project Issue Tracker")


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    if ZopeTestCase.hasProduct('PloneSoftwareCenter'):
        suite.addTest(makeSuite(TestPscTracker))
    return suite
